
/* main.cc */

#include <iostream>
#include <fstream>

#include "std.h"
#include "conv.h"
#include "inf.h"

#include "cmm_lexer.hpp"
#include "cmm_parser.hpp"
#include "cmm_product.hpp"

#include "antlr/Token.hpp"
#include "antlr/RecognitionException.hpp"
#include "antlr/LexerSharedInputState.hpp"
#include "antlr/CommonHiddenStreamToken.hpp"
#include "antlr/TokenStreamHiddenTokenFilter.hpp"

// OPEN_NAMESPACE

/* ---------------------------------------------------------------------- */

// global variables
string inp_name = "";
string out_name = "";

#define WITH_FILTER

#ifdef WITH_FILTER
typedef antlr::TokenStreamHiddenTokenFilter CmmFilter;
typedef antlr::CommonHiddenStreamToken CmmToken;

CmmLexer  * lexer  = NULL;
CmmFilter * filter = NULL;
CmmParser * parser = NULL;
#endif

/* ---------------------------------------------------------------------- */

#ifdef WITH_FILTER
void open_parser (std::istream & stream, string file_name, int line)
{
   assert (parser == NULL);
   #if 0
   set_input_name (file_name);
   #endif

   lexer = new CmmLexer (stream);
   // lexer->LocalInit (); // important
   lexer->setFilename (file_name);
   if (line != 0) lexer->setLine (line);
   lexer->setTokenObjectFactory (CmmToken::factory);

   filter = new CmmFilter (*lexer);

   filter->hide (CmmLexer::WS);
   filter->hide (CmmLexer::SL_COMMENT);
   filter->hide (CmmLexer::ML_COMMENT);

   parser = new CmmParser (*filter);
   parser->setFilename (file_name);
}

void close_parser ()
{
   assert (parser != NULL);

   delete parser;
   delete filter;
   delete lexer;

   parser = NULL;
   filter = NULL;
   lexer = NULL;
}
#endif

/* ---------------------------------------------------------------------- */

extern "C" int main (int argc, char * argv [])
{
   int err_code = 0;
   set_prog_name ("Cmm");

   // debug
   #if 1 // #ifndef NDEBUG
      if (argc == 1)
      {
         #if 1
            // #define NAME "simple"
            #define NAME "gtk1"
            // #define NAME "gtk2"

            // #define NAME "qt"
            // #define NAME "inv"
            // #define NAME "motif"
            // #define NAME "string"
            // #define NAME "iostream"
            // #define NAME "map"
            // #define NAME "queue"

            static char * tab [] =
            {"",
             "./_output/headers/" NAME ".lst",
             "-o",
             "./_output/headers/" NAME ".dump",
             // "--no-line-numbers",
             // "--comments"
             };
         #endif

         #if 0
            #define NAME "TString"

            static char * tab [] =
            {"",
             "_output/root/" NAME ".h",
             "-o",
             "_output/root/" NAME ".dump",
             // "--no-line-numbers",
             // "--comments"
             };
          #endif

         argc = sizeof (tab) / sizeof (tab[0]);
         argv = tab;
      }
   #endif

   #if 0
   // arguments
   cmm.line_numbers = true;
   cmm.comments = false;
   #endif

   bool o_opt = false;

   for (int i = 1; i < argc; i++)
   {
      string txt = argv [i];

      if (o_opt)
      {
         if (txt[0] == '-')
            error ("Output file name expected");
         if (out_name != "")
            error ("Output file name already defined");
         out_name = txt;
         o_opt = false;
      }

      else if (txt == "-o" || txt == "--output")
         o_opt = true;

      #if 0
      else if (txt == "-l")
         cmm.line_numbers = ! cmm.line_numbers;

      else if (txt == "--line-numbers")
         cmm.line_numbers = true;

      else if (txt == "--no-line-numbers")
         cmm.line_numbers = false;

      else if (txt == "-c")
         cmm.comments = ! cmm.comments;

      else if (txt == "--comments")
         cmm.comments = true;

      else if (txt == "--no-comments")
         cmm.comments = false;

      else if (txt == "-v")
         cmm.verify = ! cmm.verify;

      else if (txt == "--verify")
         cmm.verify = true;

      else if (txt == "--no-verify")
         cmm.verify = false;
      #endif

      else
      {
         if (txt[0] == '-')
            error ("Unknown option");
         if (inp_name != "")
            error ("Input file name already defined");
         inp_name = txt;
      }
   }

   if (inp_name == "") error ("Missing input file name");
   // if (out_name == "") error ("Missing output file name");

   // if (line_numbers)
   //    comments = false;

   // if (out_dir != "" && out_dir [out_dir.length()] != "/")
   //    out_dir = out_dir + '/';

   // cmm.set_input_name (inp_name);

   try
   {
      std::cerr << get_prog_name () << ": " << inp_name << " -> " << out_name << std::endl;

      std::fstream inp (inp_name.c_str (), std::ios::in);
      if (! inp.good ())
         error ("Cannot open input file: " + inp_name);

      #ifdef WITH_FILTER
      open_parser (inp, inp_name, 0);
      #else
      CmmParser * parser = CmmParser::open_parser (inp, inp_name, 0);
      #endif

      try
      {
         CmmDeclSect * d = parser->declaration_list ();
         // CmmIfStat * d = parser->if_stat ();
         // CmmCompoundStat * d = parser->compound_stat ();
         parser->match (antlr::Token::EOF_TYPE);
         // parser->CheckEnd ();

         if (out_name != "")
         {
            CmmProduct cmm;

            cmm.print_open (out_name);
            #if 0
            cmm.print_info (inp_name, out_name);
            #endif

            cmm.send_declaration_list (d);
            // cmm.send_if_stat (d);
            // cmm.send_compound_stat (d);

            #if 0
            if (cmm.comments) cmm.print_cond_eol ();
            #endif
            if (err_code == 0) cmm.print_close ();
         }
      }
      catch (antlr::RecognitionException & e)
      {
          show_error (e.getMessage (), e.getFilename (), e.getLine (), e.getColumn ());
          err_code = 1;
      }
      catch (antlr::ANTLRException & e)
      {
         antlr::LexerSharedInputState state =
         #ifdef WITH_FILTER
         lexer->getInputState ();
         #else
         parser->lexer->getInputState ();
         #endif
         show_error (e.getMessage (), state->filename, state->line, state->column);
         err_code = 1;
      }
      catch (std::exception & e)
      {
         show_error (e.what());
         err_code = 1;
      }

      #ifdef WITH_FILTER
      close_parser ();
      #else
      parser->close_parser ();
      #endif
   }
   catch (std::exception & e)
   {
      show_error (string ("exception: ") + e.what ());
      err_code = 1;
   }

   if (err_code == 0)
      show_info ("O.K.");

   return err_code;
}

// CLOSE_NAMESPACE
