
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import antlr.collections.AST;

class Decl extends Tool
{

  private FieldInfo defineField (StructInfo target,
                                 String name,
                                 TypeInfo type)
  {
     FieldInfo f = target.lookupField (name);

     if (f == null)
     {
        // create new field
        f = target.createField (name);
        f.type = type;
     }
     else
     {
        // check type of existing field
        if (! type.is_derived_from (f.type))
           error ("Field " + name + " already defined with different type " +
                  "(" + f.type.ident + " <> " + type.ident + ")");
     }

     return f;
  }

  private FieldInfo defineRealField (StructInfo target,
                                     String name,
                                     TypeInfo type)
  {
     FieldInfo f = defineField (target, name, type);
     f.is_real = true;
     return f;
  }

  private void defineEnumItem (EnumInfo target,
                               String name)
  {
     if (! target.containsItem (name))
        target.createItem (name);
  }

  private void defineConst (StructInfo target,
                            String name,
                            int value)
  {
     FieldInfo f = defineField (target, name, info.int_type);
     f.is_const = true;
     f.const_value = "" + value;
  }

  private void defineCounter (StructInfo target,
                               String name)
  {
     FieldInfo f = defineField (target, name, info.int_type);
  }

  private void defineMethod (StructInfo target,
                             String name,
                             String result_decl,
                             String params_decl)
  {
     MethodInfo m = new MethodInfo ();
     m.ident = name;
     m.result_decl = result_decl;
     m.params_decl = params_decl;
     target.addMethod (m);
  }

  private void defineConversion (TypeInfo type, TypeInfo super_type)
  {
     String name = info.conv_func_prefix + type.type_name ();

     if (! type.hasMethod (name))
     {
        MethodInfo m = new MethodInfo ();
        m.ident = name;
        m.result_decl = type.inner_decl ();
        m.params_decl = "";
        m.virtual_func = true;
        m.return_value = "this";
        type.addMethod (m);
     }

     if (! super_type.hasMethod (name))
     {
        MethodInfo m = new MethodInfo ();
        m.ident = name;
        m.result_decl = type.inner_decl ();
        m.params_decl = "";
        m.virtual_func = true;
        m.return_value = "NULL";
        super_type.addMethod (m);
     }
  }

  private void defineCustomConversion (TypeInfo type)
  {
     String name = info.conv_func_prefix + type.type_name ();

     if (! type.hasMethod (name))
     {
        MethodInfo m = new MethodInfo ();
        m.ident = name;
        m.result_decl = type.inner_decl ();
        m.params_decl = "";
        m.virtual_func = true;
        m.return_value = "this";
        type.addMethod (m);
     }

     // second method insert into basic type
     if (! info.isBasicTypeMethod (name))
     {
        MethodInfo m = new MethodInfo ();
        m.ident = name;
        m.result_decl = type.inner_decl ();
        m.params_decl = "";
        m.virtual_func = true;
        m.return_value = "NULL";
        info.addBasicTypeMethod (m);
     }
  }

  private void defineValue (StructInfo target,
                            String name,
                            String value)
  {
     if (value.equals ("true") || value.equals ("false"))
     {
        // boolean field
        defineField (target, name, info.bool_type);
     }
     else
     {
        FieldInfo f = target.lookupField (name);
        if (f == null)
        {
           // create new enumeration type
           EnumInfo t = info.lookupEnum (target.ident + "_" + name);
           f = defineField (target, name, t);
        }
        // enumeration field
        defineEnumItem (f.type.as_enum(), value);
     }
  }

  private SequenceInfo defineSequence (StructInfo target,
                                       String name,
                                       StructInfo type)
  {
     SequenceInfo s = target.lookupSequence (name);

     if (s == null)
     {
        // create new sequence
        s = target.createSequence (name);
        s.owner = target;
        s.element = type;

        // choose suffix
        s.suffix = "";
        if (name.length () > 0 && ! name.equals ("item"))
        {
            s.suffix = Character.toUpperCase (name.charAt (0)) +
                       name.substring (1);
        }

        if (! type.hasAnchor (s.ident)) // only one 'next' variable
           type.addAnchor (s); // list of owned
     }
     else
     {
        // check element type
        if (s.element != type)
           error ("Sequence " + target.ident +
                  " already defined with different type " +
                  "(" + s.element.ident + " <> " + type.ident + ")");
     }

     return s;
  }

  private void definePlace (StructInfo target,
                            String name,
                            TypeInfo type)
  {
     FieldInfo f = target.createPlace (name);
     f.type = type;
  }

  private void checkCompatibility (TypeInfo target, TypeInfo source)
  {
     if (target != source)
     {
        boolean compatible = false;

        if (target instanceof StructInfo &&
            source instanceof StructInfo )
        {
           StructInfo t = (StructInfo) source;
           while (t != null && t != target)
              t = t.from;
           if (t != null)
              compatible = true;
        }

        if (! compatible)
           error ("Incompatible types: " + target.ident + " and " + source.ident);

        defineConversion (source, target);
     }
  }

  /* location and notes */

  private int allocateNote (StructInfo type)
  {
     if (info.notes || info.locations)
     {
        info.note_decl_cnt ++;
        if (info.fixed_notes)
        {
           int inx;
           if (info.distinct_notes)
              inx = info.note_decl_cnt;
           else
              inx = type.note_count + 1;

           if (info.locations)
              definePlace (type, "location_" + inx, info.int_type);

           if (info.notes)
              definePlace (type, "start_note_" + inx, info.string_type);

           if (info.notes && ! info.only_start_notes)
              definePlace (type, "stop_note_" + inx,  info.string_type);

           return inx;
        }
        else
        {
           type.note_count ++;
           return type.note_count;
           // 0 ... note not assigned
           // 1 ... first note, stored in note [0]
           // 2 ... second note, stored in note [1]
        }
     }
     else return 0;
  }

  private void location_and_notes (BasicNode item, StructInfo type)
  {
     item.note_index = allocateNote (type);
     item.allocated = true;
  }

  private void cond_location_and_notes (BasicNode item, StructInfo type)
  {
     if (! item.allocated)
        location_and_notes (item, type);
     else
        info.note_decl_cnt ++;
  }

  private void another_location_and_notes (BasicNode item, StructInfo type)
  {
     if (item.another == null)
     {
        item.another = new BasicNode ();
        location_and_notes (item.another, type);
     }
     else
        info.note_decl_cnt ++;
  }

  /* grammar, lexer and parser options */

  public void grammarOption (String name, String value)
  {
     super.grammarOption (name, value);

     // store name and value into dictionary
     DictItem2 item = new DictItem2 ();
     item.ident = name;
     item.value = value;
     info.grammar_repository.store (item);
  }

  public void lexerOption (String name, String value)
  {
     super.lexerOption (name, value);

     DictItem2 item = new DictItem2 ();
     item.ident = name;
     item.value = value;
     info.lexer_repository.store (item);
  }

  public void parserOption (String name, String value)
  {
     super.parserOption (name, value);

     DictItem2 item = new DictItem2 ();
     item.ident = name;
     item.value = value;
     info.parser_repository.store (item);
  }

  /* enum */

  private EnumInfo act_enum;

  public void openEnum (String name)
  {
     super.openEnum (name);
     act_enum = info.createEnum (name);
  }

  public void closeEnum ()
  {
     super.closeEnum ();
     act_enum = null;
  }

  public void enumItem (String name)
  {
     super.enumItem (name);
     act_enum.createItem (name);
  }

  /* structures */

  private StructInfo act_struct;

  public void openStruct(String type_name)
  {
     super.openStruct (type_name);
     act_struct = info.findStruct (type_name);
  }

  public void closeStruct ()
  {
     super.closeStruct ();
     act_struct = null;
  }

  public void structField (String name, String type, boolean pointer)
  {
     super.structField (name, type, pointer);

     FieldInfo f = act_struct.createField (name);

     TypeInfo t = info.getType (type);
     if (t == null)
        t = info.newTypeDef (type, pointer); // !?

     f.type = t;
  }

  /* literals */

  public void literal (String name, String value)
  {
     super.literal (name, value);
     info.storeLiteral (name, value);
  }

  /* external tokens */

  public void externalToken (String name)
  {
     super.externalToken (name);
     info.storeExternalToken (name);
  }

  /* terminal */

  private void check_terminal (String value)
  {
     if (isIdentifier (value))
        info.checkToken (value);
     else
        info.checkLiteral (value);
  }

  public void terminal (String value)
  {
     super.terminal (value);
     check_terminal (value);
     if (! inside_predicate ())
        location_and_notes (current_terminal, rule.struct_type);
  }

  /* rule */

  public void openRule (String name)
  {
     super.openRule (name);
     rule.variable_name = info.variable_prefix + "result";

     if (rule.mode == NEW_MODE ||
         rule.mode == MODIFY_MODE)
         rule.rule_ready = true; // result variable is ready

     if (rule.struct_type != null && rule.struct_type.custom_conv)
        defineCustomConversion (rule.struct_type);
  }

  public void addTag (String name, String value)
  {
     super.addTag (name, value);
     defineValue (rule.struct_type, name, value);
  }

  public void addStore (String name, String type_name)
  {
     super.addStore (name, type_name);

     if (rule.mode == MODIFY_MODE)
        error ("Cannot use <store> option in rule with <modify> specification");

     defineField (rule.struct_type, rule.store_name, rule.store_type);

     // already done by Scanner
     // StructInfo t = info.findStruct (type_name);
     // defineField (rule.struct_type, name, t);
     // rule.store_redefinition = name;
     // rule.store_type = t;
  }

  /* binary */

  public void binaryBranch (String branch)
  {
     super.binaryBranch (branch);

     rule.addVariable (info.temp_prefix + "branch",
                       product_ns + info.type_prefix + info.binary_expr_type + " *");

     rule.addVariable (info.temp_prefix + "right",
                       product_ns + info.type_prefix + info.expr_type + " *");

     StructInfo expr_type = info.findStruct (info.expr_type);
     RuleNode r = info.findRule (branch);
     defineConversion (r.type, expr_type);
     rule.binary_sub_rule = branch;

     if (info.locations)
        rule.addVariable (info.temp_prefix + "location", info.location_type);

     if (info.notes)
     {
        rule.addVariable (info.temp_prefix + "start_note", info.note_type);

        if (! info.only_start_notes)
           rule.addVariable (info.temp_prefix + "stop_note", info.note_type);
     }
  }

  public void binaryItem (String key, String selector_value,
                          String middle_branch, String middle_key)
  {
     super.binaryItem (key, selector_value, middle_branch, middle_key);

     check_terminal (key);

     StructInfo expr_type = info.findStruct (info.expr_type);
     StructInfo bin_type  = info.findStruct (info.binary_expr_type);

     defineConversion (bin_type, expr_type);

     defineValue (bin_type, info.binary_expr_selector, selector_value);
     defineRealField (bin_type, info.binary_expr_left, expr_type);
     defineRealField (bin_type, info.binary_expr_right, expr_type);

     cond_location_and_notes (info.binary_node, bin_type);

     if (! empty (middle_branch))
     {
        check_terminal (middle_key);
        defineRealField (bin_type, info.binary_expr_middle, expr_type);
        another_location_and_notes (info.binary_node, bin_type);
     }
  }

  /* unary */

  public void unary (String key, String selector_value, String branch)
  {
     super.unary (key, selector_value, branch);

     rule.unary_selector = selector_value;
     rule.addVariable (info.temp_prefix + "param",
                       product_ns + info.type_prefix + info.expr_type + " *");

     check_terminal (key);
     rule.unary_key = key;

     rule.unary_sub_rule = branch;

     StructInfo expr_type = info.findStruct (info.expr_type);
     StructInfo unary_type = info.findStruct (info.unary_expr_type);

     RuleNode r = info.findRule (branch);
     defineConversion (r.type, expr_type);

     defineValue (unary_type, info.unary_expr_selector, selector_value);
     defineRealField (unary_type, info.unary_expr_param, expr_type);

     cond_location_and_notes (info.unary_node, unary_type);
  }

  /* blocks */

  public void openBlock ()
  {
     super.openBlock ();
  }

  public void closeBlock ()
  {
     super.closeBlock ();

     if (! inside_predicate ())
     {
        // result must be assigned in all inner alternatives
        // (see closeAlternative)

        // set above ready flag
        if (block.above == null)
           rule.rule_ready = true;
        else
           block.above_alt.alt_ready = true;
     }
  }

  /* alternatives */

  public void openAlternative ()
  {
     super.openAlternative ();

     if (! inside_predicate ())
     {
        // set ready flag
        if (block.above == null)
           alt.alt_ready = rule.rule_ready;
        else
           alt.alt_ready = block.above_alt.alt_ready;
     }
  }

  public void closeAlternative ()
  {
     super.closeAlternative ();

     if (! inside_predicate ())
     {
        if (! alt.alt_ready)
           if (! info.anonymous_fields) // !?
              error ("End of alternative without assigned result");
     }
  }

  /* semantic predicate */

  public void semanticPredicate (String name, boolean no_params)
  {
     super.semanticPredicate (name, no_params);
  }

  /* nonterminals */

  private void check_ready ()
  {
     if (! alt.alt_ready)
        if (! info.anonymous_fields) // !?
           error ("No result assigned for this alternative");
  }

  private void set_result (TypeInfo t)
  {
     checkCompatibility (rule.type, t); // assignment to rule result
     if (alt.alt_ready)
        error ("Result already assigned for this alternative");
     alt.alt_ready = true;
  }

  public void nonterminal (int kind, String label, String name)
  {
     super.nonterminal (kind, label, name);

     if (! inside_predicate ())
     {
        RuleNode r = info.findRule (name); // sub-rule
        int m = r.mode;  // sub-rule mode
        TypeInfo t = r.type; // sub-rule type

        t.used_as_field = true;

        if (r.mode == PRIMITIVE_MODE)
           location_and_notes (current, rule.struct_type);

        // check sub-rule mode
        if (kind == CALL)
        {
           if (m != MODIFY_MODE)
              error ("Subrule with <modify> specification required");
        }
        else
        {
           if (m == MODIFY_MODE)
              error ("Subrule with <modify> specification is not allowed here");
        }

        if (kind == CHOOSE2)
        {
           if (r.store_type == null)
              error ("Subrule with <store> option required");
        }
        else
        {
           if (r.store_type != null)
              error ("Subrule with <store> option is not allowed here");
        }

        if (kind == LABELED)
        {
           check_ready ();
           current.field = defineRealField (rule.struct_type, label, t);
           current.variable = rule.addVariable (info.temp_prefix + label, t.decl());

           if (t instanceof StructInfo)
              alt.addRefKey (label);
           else
              alt.addStrKey (label);
        }
        else if (kind == INSERT || kind == ADD)
        {
           check_ready ();
           current.variable = rule.addVariable ("item", t.decl());
           current.sequence = defineSequence (rule.struct_type, label, t.as_struct());

           if (rule.loop_ref != null &&
               rule.loop_ref != current.sequence)
              error ("Loop with two variables"); // !?

           alt.loop = true;

           block.loop = true;

           rule.loop_ref = current.sequence;
           rule.loop = true;
        }
        else if (kind == ASSIGN) // !?
        {
           current.variable = rule.addVariable (info.local_prefix + name, t.decl());
        }
        else if (kind == CALL)
        {
           check_ready ();
           checkCompatibility (t, rule.type); // assignment to sub-rule parameter
        }
        else if (kind==RESULT)
        {
           set_result (t); // assignment to rule result
        }
        else if (kind == SELECT)
        {
           set_result (t); // assignment to rule result
           alt.selection_rule = r;
           defineConversion (t, rule.type);
        }
        else if (kind == CHOOSE1)
        {
           if (rule.choose_compatible_type)
              checkCompatibility (rule.type, t); // assignment to rule result
           else
              rule.addVariable (info.variable_prefix + "tmp", t.decl());
           rule.keep_rule = r;
           // store data for ParserOutput
           current.choose_compatible_type = rule.choose_compatible_type;
        }
        else if (kind == CHOOSE2)
        {
           // !?
           set_result (t);
           checkCompatibility (r.store_type, rule.keep_rule.type); // parameter assignment
           alt.selection_rule = r;
           defineConversion (t, rule.type);
           // store data for ParserOutput
           current.choose_compatible_type = rule.choose_compatible_type;
        }
        else if (kind == ANONYMOUS)
        {
           error ("Missing variable identifier");
        }
        else error ("Internal error");
     }
  }

  /* options */

  public void invokeOption (String name)
  {
     super.invokeOption (name);
     defineMethod (rule.struct_type, name, "void", info.parser_class + " * parser");
  }

  public void executeOption (String name, boolean no_params)
  {
     super.executeOption (name, no_params);

     if (! info.ignore_execute_options)
     {
        // defineMethod (rule.struct_type, name, "void", info.parser_class + " * parser");
     }
  }

  public void redudantOption ()
  {
     super.redudantOption ();
  }

  public void locationOption (String name)
  {
     super.locationOption (name);

     int inx; // index of previous location
     if (info.distinct_notes)
        inx = info.note_decl_cnt - 1;
     else
        inx = rule.struct_type.note_count - 1;

     defineConst (rule.struct_type, name, inx);
  }

  private void no_predicate ()
  {
     if (inside_predicate ())
        error ("Invalid option inside predicate");
  }

  public void setOption (String name, String value)
  {
     super.setOption (name, value);
     no_predicate ();
     check_ready ();

     defineValue (rule.struct_type, name, value);
     alt.addOrdKey (name, value);
  }

  public void incOption (String name)
  {
     super.incOption (name);
     no_predicate ();
     check_ready ();

     defineCounter (rule.struct_type, name);
     alt.addIncKey (name);
  }

  public void silentOption ()
  {
     super.silentOption ();
     // no_predicate ();
  }

  public void additionalOption (String name)
  {
     super.additionalOption (name);
     no_predicate ();

     RuleNode subrule = info.findRule (name);
     defineConversion (subrule.type, rule.type);
  }

  public void styleOption (int style)
  {
     super.styleOption (style);
     no_predicate ();
  }

}
