
import antlr.collections.AST;

class Tool extends Common
{

  public static boolean first_pass = true;

  public RuleNode rule;
  public BlockNode block;
  public AltNode alt;
  public TerminalNode current_terminal;
  public NontermNode current;

  public SourceNode last_source_node;
  private int predicate_count;
  private int redudant_count;

  /* messages */

  public void rememberPosition (SourceNode node)
  {
     last_source_node = node;
  }

  private String location ()
  {
     String txt = "";

     if (last_source_node != null)
        txt = input_file_name + ":" + last_source_node.line_number + ": ";
     else
        txt = input_file_name + ": ";

     return txt;
  }

  private String description ()
  {
     String txt = "";

     if (rule != null)
     {
        txt = " (rule: " + rule.name;

        if (current != null)
        {
           txt = txt + ", nonterminal: ";
           if (current.kind == LABELED) txt = txt + current.label + ":";
           if (current.kind == INSERT) txt = txt + current.label + "<-";
           if (current.kind == ADD) txt = txt + "<add> ";
           if (current.kind == ASSIGN) txt = txt + current.label + "=";
           if (current.kind == CALL) txt = txt + "<call> ";
           if (current.kind == RESULT) txt = txt + "<result> ";
           txt = txt + current.subrule;
        }

        txt = txt + ")";
     }

     return txt;
  }

  public void error (String s)
  {
     s = location () + "EG Error " + description () + ": " + s;
     err (s);
  }

  public void warning (String s)
  {
     s = location () + "EG Warning " + description () + ": " + s;
     printErrInfo (s);
  }

  public void trace (String s)
  {
     s = location () + "EG Trace "  + description () + ": " + s;
     printInfo (s);
  }

  /* rules */

  public final void startRule (RuleNode node, String name)
  {
     last_source_node = node;
     rule = node;

     if (first_pass)
     {
         rule.name = name;
         info.insertRule (rule);
     }

     openRule (name);
  }

  public final void onRuleParam (int mode,
                                 String type_name)
  {
     // store mode
     if (first_pass)
        rule.mode = mode;

     if (info.anonymous_fields && empty (type_name))
        type_name = rule.name;

     ruleParam (mode, type_name);
  }

  public final void stopRule ()
  {
     closeRule ();
     rule = null;
  }

  public void onPrimitiveItem (PrimitiveNode node, String branch)
  {
     if (first_pass)
     {
        node.branch = branch;
        rule.insertPrimitiveItem (node);
     }

     primitiveItem (branch);
  }

  public void onBinaryItem (BinaryNode node,
                            String key,
                            String selector_value,
                            String middle_branch,
                            String middle_key)
  {
     if (first_pass)
     {
        node.key = key;
        node.selector_value = selector_value;
        rule.insertBinaryItem (node);
        node.middle_branch = middle_branch;
        node.middle_key = middle_key;
     }
     binaryItem (key, selector_value, middle_branch, middle_key);
  }

  /* blocks */

  public final void startBlock (BlockNode node)
  {
     last_source_node = node;

     // link to above block
     if (first_pass)
     {
        if (block == null)
           rule.normal_block = node; // top level block
        else
        {
           node.above = block; // link to above block
           alt.insertItem (node); // add into list
        }
     }

     block = node; // set 'block'
     block.above_alt = alt; // save 'alt'

     if (first_pass)
     {
        if (alt != null)
        {
           alt.block_count ++;
           block.index = alt.block_count;
        }
        else
        {
           block.index = 1; // top level block
        }

        if (! inside_predicate ())
           if (rule.mode == SELECT_MODE || rule.mode == CHOOSE_MODE)
              if (! is_select_block () &&
                  ! is_main_choose_block () &&
                  ! is_choose_block ())
                 error ("Invalid block in select/choose rule");

        // set 'choose_compatible_type' (also for mark == '?')
        if (block.mark != ' ' && is_choose_block ())
           rule.choose_compatible_type = true;
     }

     openBlock ();
  }

  public final void stopBlock ()
  {
     closeBlock ();

     // restore 'alt' and 'block'
     alt = block.above_alt;
     block = block.above;
  }

  public boolean is_select_block ()
  {
      return rule.mode == SELECT_MODE &&
             block.above == null;
  }

  public boolean is_main_choose_block ()
  {
      return rule.mode == CHOOSE_MODE &&
             block.above == null;
  }

  public boolean is_choose_block ()
  {
     return rule.mode == CHOOSE_MODE &&
            block.above != null &&
            block.above.above == null &&
            block.index == 1; // first block on second level
  }

  /* alternatives */

  public final void startAlternative (AltNode node)
  {
     last_source_node = node;
     alt = node;

     // set link to above block
     alt.above = block;

     // block count and index
     if (first_pass)
     {
        block.alt_count ++;
        alt.index = block.alt_count;

        if (is_main_choose_block ())
           if (alt.index != 1)
              error ("Invalid alternative in choose rule");

        // link
        block.insertAlt (alt);
     }

     alt.item_redudant = false;
     alt.next_redudant = false;

     openAlternative ();
  }

  public final void stopAlternative ()
  {
     closeAlternative ();

     // restore 'alt'
     alt = null;
  }

  /* alternatives */

  public final void startAltItem (AST ast)
  {
     if (alt.item_redudant)
        redudant_count ++;

     // link
     /*
     if (first_pass)
     {
        AST sub = ast.getFirstChild ();
        // System.out.println (sub.getClass () + " " + sub.getText ());
        SourceNode item = (SourceNode) sub;

        alt.insertLast (item);
     }
     */
  }

  public final void stopAltItem ()
  {
     if (alt.item_redudant)
        redudant_count --;

     alt.item_redudant = alt.next_redudant;
     alt.next_redudant = false;
  }

  public void onRedudantOption ()
  {
     alt.next_redudant = true;
     redudantOption ();
  }

  /* syntactic predicates */

  public final void startSyntacticPredicate ()
  {
     predicate_count ++;
     openPredicate ();
  }

  public final void stopSyntacticPredicate ()
  {
     closePredicate ();
     predicate_count --;
  }

  public boolean inside_predicate ()
  {
     return predicate_count > 0;
  }

  public boolean inside_redudant ()
  {
     return redudant_count > 0;
  }

  /* semantic predicates */

  public void onSemanticPredicate (SemanticNode node,
                                   String name,
                                   boolean no_params)
  {
     if (first_pass)
     {
        node.name = name;
        node.no_params = no_params;

        alt.insertItem (node);
     }

     semanticPredicate (name, no_params);
  }

  /* silent */

  public void onSilentOption ()
  {
     if (first_pass)
        alt.silent = true;

     silentOption ();
  }

  /* terminal */

  public void onTerminal (TerminalNode node, String name)
  {
     last_source_node = node;

     if (first_pass)
     {
        node.term_name = name;
        alt.insertItem (node);
     }

     current_terminal = node;
     terminal (name);
     current_terminal = null;
  }

  /* nonterminal */

  public final void onNonterminal (NontermNode node,
                                   int kind,
                                   String label,
                                   String name)
  {
     last_source_node = node;

     current = node;

     if (info.anonymous_fields)
     {
        if (kind == ANONYMOUS)
        {
           kind = LABELED;
           label = name; // !? duplicated identifier
        }
     }

     if (kind == ADD) label = "item"; // default sequence name

     // !?
     if (! inside_predicate ())
     {
        if (is_select_block ())
           kind = SELECT;
        if (is_main_choose_block ())
           kind = CHOOSE1;
        if (is_choose_block ())
           kind = CHOOSE2;
        if (rule.mode == RETURN_MODE && block.above == null && current.index == 1)
           kind = RESULT;
     }

     // DO NOT USE if (first_pass), SET VALUE AGAIN
     current.kind = kind;
     current.label = label;
     current.subrule = name;

     if (first_pass)
     {
        alt.nonterm_count ++;
        current.index = alt.nonterm_count;

        alt.insertItem (current);
     }

     if (first_pass)
     {
        if (kind == SELECT | kind == CHOOSE1 | kind == CHOOSE2)
           if (current.index != 1)
              error ("Additional nonterminal in select/choose rule");
     }

     nonterminal (kind, label, name);

     current = null;
  }

  /* options */

  public void onSetOption (SetOptionNode node,
                           String name,
                           String value)
  {
     if (first_pass)
     {
        node.name = name;
        node.value = value;
        alt.insertItem (node);
     }

     setOption (name, value);
  }

  public void onIncOption (IncOptionNode node,
                           String name)
  {
     if (first_pass)
     {
        node.name = name;
        alt.insertItem (node);
     }

     incOption (name);
  }

  public void onInvokeOption (InvokeOptionNode node,
                              String name)
  {
     if (first_pass)
     {
        node.name = name;
        alt.insertItem (node);
     }

     invokeOption (name);
  }

  public void onExecuteOption (ExecuteOptionNode node,
                               String name,
                               boolean no_params)
  {
     if (first_pass)
     {
        node.name = name;
        node.no_params = no_params;
        alt.insertItem (node);
     }

     executeOption (name, no_params);
  }

  /* actions */

  public void openStruct (String type_name) { }
  public void closeStruct () { }
  public void structField (String name, String type, boolean pointer) { }
  public void structSequence (String name, String type) { }

  public void addSuper (String type_name, String super_name) { }
  public void addTag (String name, String value) { }
  public void addCustom (String key) { }
  public void addStore (String name, String type_name) { }

  public void openEnum (String name) { }
  public void closeEnum () { }
  public void enumItem (String name) { }

  public void literal (String name, String value) { }
  public void externalToken (String name) { }

  public void option (String name, String value) { }
  public void grammarOption (String name, String value) { }
  public void lexerOption (String name, String value) { }
  public void parserOption (String name, String value) { }
  public void extension (String name, String value) { }

  public void publicOption () { }
  public void privateOption () { }

  public void openRule (String name) { }
  public void closeRule () { }
  public void ruleParam (int mode, String type_name) { }

  public void openBinary () { }
  public void closeBinary () { }
  public void binaryBranch (String branch) { }
  public void binaryItem (String key, String selector_value, String middle_branch, String middle_key) { }

  public void unary (String key, String selector_value, String branch) { }

  public void openPrimitive () { }
  public void closePrimitive () { }
  public void identifierOption () { }
  public void primitiveItem (String branch) { }

  public void openBlock () { }
  public void closeBlock () { }

  public void openAlternative () { }
  public void closeAlternative () { }

  public void openPredicate () { }
  public void closePredicate () { }

  public void terminal (String name) { }
  public void nonterminal (int kind, String label, String name) { }

  public void semanticPredicate (String name, boolean no_params) { }

  public void setOption (String name, String value) { }
  public void incOption (String name) { }
  public void silentOption () { }
  public void additionalOption (String name) { }

  public void invokeOption (String name) { }
  public void executeOption (String name, boolean no_params) { }
  public void redudantOption () { }
  public void locationOption (String name) { }
  public void styleOption (int style) { }

  /* initialization */

  public void onInit () { }
  public void onFinish () { }
}
