
/* envio.cc */

#include "envio.h"

#include "ioexception.h"
#include "str.h" // const lf, ...
#include "conv.h" // template functions StrToNum, NumToStr
#include "inf.h" // function info

#include <string.h>  // function memcpy

#define DUPL_OUTPUT

OPEN_NAMESPACE

/****************************** TOKEN TYPES *******************************/

string TokenKindToStr (TTokenKind kind)
{
   static string table [] =
   {
      "StartToken",

      "IdentToken",
      "NumberToken",
      "CharToken",
      "StringToken",
      "SeparatorToken",

      "SpaceToken",
      "CommentToken",
      "LineContToken",
      "EolToken",

      "TextLineToken",

      "UnusedDirectiveToken",
      "IncludeDirectiveToken",
      "ProcessedDirectiveToken",

      "StopToken",
      "FinalToken",
      "EofToken"
   };

   const int lim = sizeof (table) / sizeof (table [0]);
   assert (EofToken == lim-1);

   string result;

   if (kind >= 0 && kind < lim)
      result = table [kind];
   else
      result = "TokenKind (" + IntToStr (kind) + ")";

   return result;
}

/************************** INCLUDE DIRECTORIES ***************************/

TInclDir::TInclDir () :
   name (),
   kind (normal_incl),
   next (NULL)
{
}

TInclDirectories::TInclDirectories () :
   first_dir (NULL),
   last_dir (NULL)
{
}

TInclDirectories::~TInclDirectories ()
{
   TInclDir * item = first_dir;
   while (item != NULL)
   {
      TInclDir * tmp = item->next;
      delete item;
      item = tmp;
   }
   first_dir = NULL;
   last_dir = NULL;
}

/* ---------------------------------------------------------------------- */

void TInclDirectories::InsInclDir (string name, TInclKind kind)
{
   assert (this != NULL);

   TInclDir * item = new TInclDir ();
   item->name = name;
   item->kind = kind;

   item->next = NULL;

   if (first_dir == NULL)
   {
      assert (last_dir == NULL);
      first_dir = item;
      last_dir = item;
   }
   else
   {
      assert (last_dir != NULL);
      last_dir->next = item;
      last_dir = item;
   }
}

void TInclDirectories::AddInclDir (string name, TInclKind kind)
{
   TInclDir * item = first_dir;
   while (item != NULL && item->name != name)
      item = item->next;

   if (item == NULL)
      InsInclDir (name, kind);
}

/* ---------------------------------------------------------------------- */

void TInclDirectories::CopyInclDirectoriesFrom (TInclDirectories * source)
{
   assert (source != NULL);
   TInclDir * item = source->first_dir;
   while (item != NULL)
   {
      InsInclDir (item->name, item->kind);
      item = item->next;
   }
}

/* ---------------------------------------------------------------------- */

inline bool FileExists (string file_name)
{
   ifstream f;

   f.open (file_name.c_str ());
   bool found = f.good ();
   f.close ();

   return found;
}

inline bool ItemFound (TInclDir * item, string name, bool system_file)
{
   bool result = false;

   bool allowed;

   if (system_file)
      allowed = (item->kind != user_incl);
   else
      allowed = (item->kind != system_incl);

   if (allowed)
   {
      string file_name = AddSlash (item->name) + name;

      // debug ("try " + file_name);
      if (FileExists (file_name))
      {
         // debug ("found " + file_name);
         result = true;
      }
   }

   return result;
}

/* ---------------------------------------------------------------------- */

string TInclDirectories::FindInclFile (string name,
                                       bool system_file,
                                       string include_next,
                                       bool show_incl_next)
{
   assert (this != NULL);
   string result = ""; // when not found

   if (! HasPrefix (name, '/'))
   {
      bool found = false;
      TInclDir * item = first_dir;

      while (item != NULL && ! found)
      {
         bool allowed;

         if (system_file)
            allowed = (item->kind != user_incl);
         else
            allowed = (item->kind != system_incl);

         if (allowed)
         {
            string file_name = AddSlash (item->name) + name;

            // debug ("try FindInclFile " + file_name);
            if (FileExists (file_name))
            {
               // debug ("found FindInclFile " + file_name);
               if (include_next != "")
               {
                  if (show_incl_next)
                     info ("Found " + file_name + ", switching to  " + include_next);
                  name = include_next; // search again (possibly another file name)
                  include_next = "";
               }
               else
               {
                  found = true;
                  result = file_name;
               }
            }
         }
         item = item->next;
      }
   }

   return result;
}

string TInclDirectories::LookupInclFile (string name,
                                         string dir,
                                         bool system_file,
                                         string include_next,
                                         bool show_incl_next)
{
   string result = "";

   string local_name = AddSubdir (dir, name);
   // debug ("LookupInclFile " + local_name + ", next: " + include_next);

   if (! system_file && FileExists (local_name))
   {
      if (include_next != "")
      {
         if (show_incl_next)
            info  ("Found " + local_name + ", switching to  " + include_next);
         name = include_next; // search again (possibly another file name)
         include_next = "";
         result = FindInclFile (name, system_file, include_next, show_incl_next);
      }
      else
      {
         result = local_name;
      }
   }
   else
   {
      result = FindInclFile (name, system_file, include_next, show_incl_next);
   }

   // debug ("end of LookupInclFile " + result);
   return result;
}

/************************** CONDITIONAL SYMBOLS ***************************/

TCondParam::TCondParam () :
   name (),
   multi (false),
   next (NULL)
{
}

/* ---------------------------------------------------------------------- */

TCondSymbol::TCondSymbol () :
   name (),
   value (""),
   with_params (false),
   local_param (false),
   orig_value (""),
   first_param (NULL),
   last_param (NULL),
   prev (NULL),
   next (NULL),
   above (NULL),
   FileName (),
   FileInx (0),
   LinePos (0),
   ColPos (0),
   CharPos (0)
{
}

void TCondSymbol::DeleteParameters ()
{
   TCondParam * item = first_param;
   while (item != NULL)
   {
      TCondParam * tmp = item->next;
      delete item;
      item = tmp;
   }
   first_param = NULL;
   last_param = NULL;
}

TCondSymbol::~TCondSymbol ()
{
   DeleteParameters ();

   if (above != NULL)
   {
      if (prev != NULL)
         prev->next = next;
      else
         above->first_symbol = next;

      if (next != NULL)
         next->prev = prev;
      else
         above->last_symbol = prev;

      above = NULL;
      prev = NULL;
      next = NULL;
   }
}

/* ---------------------------------------------------------------------- */

TCondSymbols::TCondSymbols () :
   first_symbol (NULL),
   last_symbol (NULL)
{
}

TCondSymbols::~TCondSymbols ()
{
   TCondSymbol * item = first_symbol;
   while (item != NULL)
   {
      TCondSymbol * tmp = item->next;
      delete item;
      item = tmp;
   }

   first_symbol = NULL;
   last_symbol = NULL;
}

/* ---------------------------------------------------------------------- */

void TCondSymbol::InsParam (string name, bool multi)
{
   assert (this != NULL);

   TCondParam * item = new TCondParam;
   item->name = name;
   item->multi = multi;
   item->next = NULL;

   if (first_param == NULL)
   {
      assert (last_param == NULL);
      first_param = item;
      last_param = item;
   }
   else
   {
      assert (last_param != NULL);
      last_param->next = item;
      last_param = item;
   }
}

void TCondSymbols::InsSymbol (TCondSymbol * item)
{
   assert (this != NULL);
   assert (item != NULL);

   item->prev = last_symbol;
   item->next = NULL;
   item->above = this;

   if (first_symbol == NULL)
   {
      assert (last_symbol == NULL);
      first_symbol = item;
      last_symbol = item;
   }
   else
   {
      assert (last_symbol != NULL);
      last_symbol->next = item;
      last_symbol = item;
   }
}

/* ---------------------------------------------------------------------- */

TCondSymbol * TCondSymbols::SearchSymbol (string name)
{
   assert (this != NULL);
   TCondSymbol * item = first_symbol;
   while (item != NULL && item->name != name)
      item = item->next;
   return item;
}

string TCondSymbols::RecallSymbol (string name)
{
   TCondSymbol * item = SearchSymbol (name);

   if (item == NULL)
      return "";
   else
      return item->value;
}

/* ---------------------------------------------------------------------- */

void TCondSymbol::CopyCondParametersFrom (TCondSymbol * source)
{
   assert (source != NULL);
   TCondParam * item = source->first_param;
   while (item != NULL)
   {
      InsParam (item->name, item->multi);
      item = item->next;
   }
}

void TCondSymbols::CopyCondSymbolsFrom (TCondSymbols * source)
{
   assert (source != NULL);
   TCondSymbol * item = source->first_symbol;
   while (item != NULL)
   {
      TCondSymbol * tmp = new TCondSymbol;

      tmp->name = item->name;
      tmp->value = item->value;

      tmp->FileName = item->FileName;
      tmp->LinePos = item->LinePos;
      tmp->ColPos = item->ColPos;
      tmp->CharPos = item->CharPos;

      tmp->CopyCondParametersFrom (item);

      InsSymbol (tmp);

      item = item->next;
   }
}

/* ---------------------------------------------------------------------- */

bool TCondSymbols::IsDefinedSymbol (string name)
{
   TCondSymbol * item = SearchSymbol (name);
   return item != NULL;
}

TCondSymbol * TCondSymbols::DefineNewSymbol (string name, string value)
{
   TCondSymbol * item = SearchSymbol (name);

   if (item == NULL)
   {
      item = new TCondSymbol;
      item->name = name;
      item->value = value;
      InsSymbol (item);
   }
   else
   {
      item->DeleteParameters ();
      item->value = value;
   }

   return item;
}

void TCondSymbols::DefineSymbolWithParams (string name, string params, string value)
{
   TCondSymbol * item = DefineNewSymbol (name, value);

   if (params != "")
   {
      item->with_params = true; // important

      int inx = 0;
      int len = params.length ();

      while (inx < len)
      {
         int start = inx;

         while (inx < len && params [inx] != ',')
            inx ++;

         string s = params.substr (start, inx-start);
         s = Trim (s);
         // debug ("Define Item " + name + " with parameter " + s);
         item->InsParam (s);

         if (inx < len && params [inx] == ',')
            inx ++;
      }
   }
}

void TCondSymbols::DefineSymbol (string name, string value)
{
   DefineNewSymbol (name, value);
}

void TCondSymbols::UndefineSymbol (string name)
{
   TCondSymbol * item = SearchSymbol (name);
   if (item != NULL)
      delete item;
}

/************************** ENVIRONMENT VARIABLES *************************/

TEnvVar::TEnvVar () :
   name (),
   value (),
   next (NULL)
{
}

TEnvVariables::TEnvVariables () :
   first_var (NULL),
   last_var (NULL)
{
}

TEnvVariables::~TEnvVariables ()
{
   TEnvVar * item = first_var;
   while (item != NULL)
   {
      TEnvVar * tmp = item->next;
      delete item;
      item = tmp;
   }
   first_var = NULL;
   last_var = NULL;
}

/* ---------------------------------------------------------------------- */

void TEnvVariables::InsVariable (string name, string value)
{
   // debug (PtrToStr (this) + " AddPathVar " + name + " = " + value);
   assert (this != NULL);

   TEnvVar * item = new TEnvVar ();
   item->name = name;
   item->value = value;

   if (first_var == NULL)
   {
      assert (last_var == NULL);
      first_var = item;
      last_var = item;
   }
   else
   {
      assert (last_var != NULL);
      last_var->next = item;
      last_var = item;
   }
}

TEnvVar * TEnvVariables::SearchVariable (string name)
{
   assert (this != NULL);
   TEnvVar * item = first_var;
   while (item != NULL && item->name != name)
      item = item->next;
   return item;
}

void TEnvVariables::AddVariable (string name, string value)
{
   TEnvVar * item = SearchVariable (name);
   if (item == NULL)
      InsVariable (name, value);
   else
      item->value = value;
}

string TEnvVariables::RecallVariable (string name)
{
   TEnvVar * item = SearchVariable (name);
   if (item == NULL)
      return "";
   else
      return item->value;
}

void TEnvVariables::CopyEnvVariablesFrom (TEnvVariables * source)
{
   assert (source != NULL);
   TEnvVar * item = source->first_var;
   while (item != NULL)
   {
      InsVariable (item->name, item->value);
      item = item->next;
   }
}

/* ---------------------------------------------------------------------- */

#if 0
string TEnvVariables::TranslateVariable (string name)
// resolve variable names in following strings $(variable)/relative_path
{
   string::size_type len = name.length ();

   if (len > 1 && name[0] == '$')
   {
      string key = "";
      string tail = "";

      if (name[1] == '(')
      {
         string::size_type inx = 2;
         while (inx < len && name [inx] != ')')
            inx ++;

         if (inx > 2 && inx < len)
         {
            key = name.substr (2, inx-2);
            tail = name.substr (inx+1);
         }
      }
      else
      {
         string::size_type inx = 1;
         while (inx < len && IsLetter (name [inx]))
            inx ++;

         if (inx > 1)
         {
            key = name.substr (1, inx-1);
            tail = name.substr (inx);
         }
      }

      if (key != "")
      {
         TEnvVar * item = SearchVariable (key);
         if (item != NULL)
            name = RemoveLastSlash (item->value) + tail;
      }
   }

   // debug (" result is " + name);
   return name;
}
#endif

/******************************* USED FILES *******************************/

TUsedFile::TUsedFile () :
   name (),
   next (NULL)
{
}

TUsedFiles::TUsedFiles () :
   first_file (NULL),
   last_file (NULL)
{
}

TUsedFiles::~TUsedFiles ()
{
   TUsedFile * item = first_file;
   while (item != NULL)
   {
      TUsedFile * tmp = item->next;
      delete item;
      item = tmp;
   }
   first_file = NULL;
   last_file = NULL;
}

/* ---------------------------------------------------------------------- */

void TUsedFiles::InsUsedFile (string name)
{
   assert (this != NULL);

   TUsedFile * item = new TUsedFile ();
   item->name = name;
   item->next = NULL;

   if (first_file == NULL)
   {
      assert (last_file == NULL);
      first_file = item;
      last_file = item;
   }
   else
   {
      assert (last_file != NULL);
      last_file->next = item;
      last_file = item;
   }
}

void TUsedFiles::CopyUsedFilesFrom (TUsedFiles * source)
{
   assert (source != NULL);
   TUsedFile * item = source->first_file;
   while (item != NULL)
   {
      InsUsedFile (item->name);
      item = item->next;
   }
}

/* ---------------------------------------------------------------------- */

void TUsedFiles::AddUsedFile (string name)
{
   TUsedFile * item = first_file;
   while (item != NULL && item->name != name)
      item = item->next;

   if (item == NULL)
      InsUsedFile (name);
}

/****************************** ENVIRONMENT *******************************/

TTextEnv::TTextEnv ()
{
}

TTextEnv::~TTextEnv ()
{
}

void TTextEnv::CopyFrom (TTextEnv * source)
{
   assert (this != NULL);

   CopyInclDirectoriesFrom (source);
   CopyCondSymbolsFrom (source);
   CopyEnvVariablesFrom (source);
   CopyUsedFilesFrom (source);
}

string TTextEnv::TranslatePath (string name)
// resolve symbol names and variable names in following strings $(variable)/relative_path
{
   str_size_type len = name.length ();

   if (len > 1 && name[0] == '$')
   {
      string key = "";
      string tail = "";

      if (name[1] == '(' || name[1] == '{')
      {
         char stop = (name[1] == '(') ? ')' : '}';

         str_size_type inx = 2;
         while (inx < len && name [inx] != stop)
            inx ++;

         if (inx > 2 && inx < len)
         {
            key = Substr (name, 2, inx-2);
            tail = Tail (name, inx+1);
         }
      }
      else
      {
         str_size_type inx = 1;
         while (inx < len && IsLetter (name [inx]))
            inx ++;

         if (inx > 1)
         {
            key = Substr (name, 1, inx-1);
            tail = Tail (name, inx);
         }
      }

      if (key != "")
      {
         TCondSymbol * s = SearchSymbol (key);
         if (s != NULL)
            name = RemoveLastSlash (s->value) + tail;
         else
         {
            TEnvVar * v = SearchVariable (key);
            if (v != NULL)
               name = RemoveLastSlash (v->value) + tail;
         }
      }
   }

   return name;
}

/*************************** CONDITIONAL LEVELS ***************************/

TCondLevel::TCondLevel () :
   value (false),
   done (false),
   prev (NULL),
   next (NULL)
{
}

TCondLevels::TCondLevels () :
   first_level (NULL),
   last_level (NULL),
   current_level (NULL),
   level (0)
{
}

TCondLevels::~TCondLevels()
{
   TCondLevel * item = first_level;
   while (item != NULL)
   {
      TCondLevel * tmp = item->next;
      delete item;
      item = tmp;
   }
   first_level = NULL;
   last_level = NULL;
   current_level = NULL;
   level = 0;
}

/* ---------------------------------------------------------------------- */

void TCondLevels::InsCondLevel (TCondLevel * item)
{
   assert (this != NULL);
   assert (item != NULL);

   item->prev = last_level;
   item->next = NULL;

   if (first_level == NULL)
   {
      assert (last_level == NULL);
      first_level = item;
      last_level = item;
   }
   else
   {
      assert (last_level != NULL);
      last_level->next = item;
      last_level = item;
   }
}

/* ---------------------------------------------------------------------- */

const int cond_level_max = 128; // only for loop detection
const int sub_level_max = 32; // only for loop detection

void TCondLevels::AddCondLevel (bool value)
{
   // look for unused item
   if (current_level == NULL)
      current_level = first_level;
   else
      current_level = current_level->next;

   if (current_level == NULL)
   {
      // allocate new item
      current_level = new TCondLevel;
      InsCondLevel (current_level);
   }

   current_level->value = value;
   current_level->done = false;

   level ++;
}

void TCondLevels::RemoveCondLevel ()
{
   assert (current_level != NULL);
   current_level = current_level->prev; // old item may be reused later
   level --;
}

void TCondLevels::RemoveAllLevels ()
{
   current_level = NULL;
   level = 0;
   // other items may be reused later
}

bool TCondLevels::Calculate ()
{
   bool result = true;

   if (current_level != NULL)
   {
      TCondLevel * item = first_level;
      while (item != current_level && result)
      {
         assert (item != NULL);
         if (! item->value)
            result = false;
         item = item->next;
      }

      if (result && ! item->value)
         result = false;
   }

   return result;
}

/******************************** OPTIONS *********************************/

TEnvOptions::TEnvOptions ():

   LowerCaseIdentifiers (false),
   UpperCaseIdentifiers (false),

   PascalNumbers (false),
   PascalStrings (false),

   PascalComments (false),
   SemicolonComments (false),
   HashComments (false),

   TextLines (false),

   ReportComments (false),
   ReportLineCont (false),
   ReportSpaces (false),
   ReportEol (false),

   HashDirectives (true), /* <-- */
   AtDirectives (false),
   DollarDirectives (false),

   KeepEmptySections (false),
   KeepUnusedSections (false),

   RequireSymbolDefinition (false),
   TolerateMissingEndOfLine (false),

   ReportIncludes (false),
   IgnoreIncludes (false),
   TolerateIncludes (false),

   IgnoreWarningDirectives (false),
   IgnoreErrorDirectives (false),

   IgnoreUnknownDirectives (false),

   IgnoreMacroExpansion (false),

   ShowInclDirectories (false),
   ShowInclNext (false),
   ShowIncludes (false),
   ShowDefinitions (false),
   ShowMacroExpansion (false),
   ShowExpansionDetails (false),
   ShowTokens (false)
{ }

/******************************* SUB INPUT ********************************/

TSubInput::TSubInput ():
   prev (NULL),
   next (NULL),
   title (),
   orig_include_name (),
   is_include_file (false),
   add_new_line (false),
   add_final_token (false),
   save_real_input (NULL),
   guard_state (before_guard),
   guard_level (0),
   guard_name ("")
{
}

TSubInput::~TSubInput ()
{
}

/* ---------------------------------------------------------------------- */

TStrSubInput::TStrSubInput ():
   inx (0),
   value ()
{
}

TStrSubInput::~TStrSubInput ()
{
}

void TStrSubInput::ReadBuf ()
{
   BufPos = 0;
   BufLen = 0;

   int len = value.length ();
   if (inx < len)
   {
      BufLen = len - inx;

      if (BufLen > BufSize)
         BufLen = BufSize;

      memcpy (Buf, value.c_str () + inx, BufLen);

      inx += BufLen;
   }
}

/******************************* ENV INPUT ********************************/

TEnvInput::TEnvInput ():

   first_input (NULL),
   real_input (NULL),
   input (NULL),

   sub_level (0),
   line_cont (0),

   CondTable (),
   CondValue (true),

   env (NULL),

   opt (),

   compound_separators (false)

{
   OpenSubInput (new TSubInput); // !? delete
   real_input = input;
   input->save_real_input = real_input;
}

TEnvInput::~TEnvInput ()
{
   // delete input streams
   while (input != NULL)
      CleanSubInput ();
}

/* ---------------------------------------------------------------------- */

void TEnvInput::CheckEnd ()
{
   if (input != NULL && input->prev != NULL)
      Error ("Remaining nested input");

   if (! InpEof)
      Error ("Remaining text");
}

void TEnvInput::ShowPositions ()
{
   TSubInput * temp = input;

   string msg = "";

   if (temp != NULL)
   {
      // current position
      msg = "Source code position";
      // msg += " in " + temp->GetInfo ();
      show_info (msg, RecallFileName(temp->TokenFileInx), temp->TokenLinePos, temp->TokenColPos);

      msg = temp->GetMsg (); // message from inner level
      // msg += " in " + temp->GetInfo ();
      temp = temp->prev;
   }

   // outer positions
   while (temp != NULL)
   {
      // msg += " in " + temp->GetInfo ();
      show_info (msg, RecallFileName(temp->TokenFileInx), temp->TokenLinePos, temp->TokenColPos);

      msg = temp->GetMsg (); // message from inner level
      // msg += " in " + temp->GetInfo ();
      temp = temp->prev;
   }
}

/* ---------------------------------------------------------------------- */

void TEnvInput::Init ()
{
   // original method
   TTextInput::Init ();

   CondTable.RemoveAllLevels ();
   CondValue = true;

   // do not read first character, already done by TTextInput::Init ()
}

/* ---------------------------------------------------------------------- */

void TEnvInput::SetStartToken ()
{
   TokenKind = StartToken;
   TokenTxt = "";
   TokenVal = "";
   TokenSym = 0;
}

void TEnvInput::SetSpaceToken (const string txt)
{
   TokenKind = SpaceToken;
   TokenTxt = txt;
   TokenVal = "";
   TokenSym = 0;
}

void TEnvInput::SetStringToken (const string txt)
{
   TokenKind = StringToken;
   TokenTxt = txt;
   TokenVal = "";
   TokenSym = 0;
}

void TEnvInput::SetNumberToken (const string txt)
{
   TokenKind = StringToken;
   TokenTxt = txt;
   TokenVal = "";
   TokenSym = 0;
}

void TEnvInput::SetIncludeDirectiveToken (const string txt)
{
   TokenKind = IncludeDirectiveToken;
   TokenTxt = "#include " + QuoteStr (txt) + lf;
   TokenVal = txt;
   TokenSym = 0;
}

void TEnvInput::SetUnusedDirectiveToken (const string txt)
{
   TokenKind = UnusedDirectiveToken;
   TokenTxt = txt;
   TokenVal = "";
   TokenSym = 0;
}

void TEnvInput::SetProcessedDirectiveToken ()
{
   TokenKind = ProcessedDirectiveToken;
   TokenTxt = "";
   TokenVal = "";
   TokenSym = 0;
}

/* ---------------------------------------------------------------------- */

void TEnvInput::GetSimpleToken () // %%
{
   assert (TokenKind != EofToken);
   assert (input != NULL);

   while (input->InpEof &&
          ! input->add_new_line &&
          ! input->add_final_token &&
          input != first_input)
   {
      // close include file or macro
      CloseSubInput ();

      InpCh = input->InpCh;
      InpEof = input->InpEof;
   }

   // store token position
   input->TokenFileInx = input->FileInx;
   input->TokenLinePos = input->LinePos;
   input->TokenColPos  = input->ColPos;
   input->TokenCharPos = input->CharPos;

   value_storage.reset ();
   text_storage.reset ();

   if (InpEof)                                     // END OF FILE
   {
      TokenTxt = "";
      TokenVal = "";
      TokenSym = 0;

      if (input != NULL && input->add_new_line)
      {
         TokenKind = EolToken;
         TokenTxt = CharToStr (lf);
         input->add_new_line = false;
      }
      else if (input != NULL && input->add_final_token)
      {
         TokenKind = FinalToken;
         input->add_final_token = false;
      }
      else
      {
         if (TokenKind == StopToken || TokenKind == EofToken)
            Error ("Unexpected end of source");

         if (input != NULL && input->prev != NULL)
            TokenKind = StopToken;
         else
            TokenKind = EofToken;
      }
   }

   else if (IsLetter (InpCh))                      // IDENTIFIER
      GetIdent (opt.LowerCaseIdentifiers,
                opt.UpperCaseIdentifiers);

   else if (IsDigit (InpCh))                       // NUMBER
      GetNumber (opt.PascalNumbers);

   else if (InpCh == '.')                          // NUMBER or PERIOD
      GetNumber (opt.PascalNumbers);

   else if (InpCh == quote1)                       // CHARACTER
      GetChar (opt.PascalStrings);

   else if (InpCh == quote2 )                      // STRING
      GetString (opt.PascalStrings);

   else if (InpCh == '`' && opt.TextLines)         // TEXT LINE
      GetTextLine ();

   else if (InpCh == '/')                          // COMMENT (or slash)
   {
      if (opt.PascalComments)
         GetPascalComment ();
      else
         GetComment ();
   }

   else if (InpCh == '{' && opt.PascalComments)    // COMMENT
      GetPascalComment ();

   else if (InpCh == '(' && opt.PascalComments)    // COMMENT (or asterisk)
      GetPascalComment ();

   else if (InpCh == '#' && opt.HashComments)      // COMMENT
      GetLineComment ();

   else if (InpCh == ';' && opt.SemicolonComments) // COMMENT
      GetLineComment ();

   else if (InpCh == cr || InpCh == lf)            // END OF LINE
      GetEol ();

   else if (InpCh <= ' ')                          // WHITE SPACE
   {
      assert (! InpEof);
      assert (! input->InpEof);
      GetSpace ();
   }

   else if (InpCh == backslash)                    // BACKSLASH
      GetBackslash ();

   else                                            // SEPARATOR
      GetSeparator ();

   #if 0
   if (opt.ShowExpansionDetails)
      Info ("GetSimpleToken: " +  TokenKindToStr (TokenKind) + " " + QuoteStr (TokenTxt) +
            ", next character is " + QuoteStr (CharToStr (InpCh)));
   #endif
}

/* ---------------------------------------------------------------------- */

void TEnvInput::OpenSubInput (TSubInput * item) // %%
{
   if (first_input == NULL)
   {
      assert (input == NULL);
      first_input = item;
      item->prev = NULL;
   }
   else
   {
      item->prev = input;
      input->next = item;
   }

   item->next = NULL;
   input = item;

   inp = input;
   #if 0
   if (opt.ShowExpansionDetails)
      info ("opening sub input: " + QuoteStr (item->GetInfo ()));
   #endif

   sub_level ++;
   if (sub_level > sub_level_max)
      Error ("Too many nested includes or macros");

   input->Init ();
   InpCh = input->InpCh;
   InpEof = input->InpEof; // important

   if (input->is_include_file)
   {
      real_input = input;
      OpeningInclude ();
   }

   input->save_real_input = real_input;
}

void TEnvInput::CleanSubInput ()
{
   assert (input != NULL);

   // remove input stream from list of include files
   TCharInput * tmp = input;
   input = input->prev;

   if (input == NULL)
      first_input = NULL;
   else
      input->next = NULL;

   delete tmp;

   inp = input;
   sub_level --;

   // important
   if (input == NULL)
   {
      InpCh = 0;
      InpEof = true;
      real_input = NULL;
   }
   else
   {
      InpCh = input->InpCh;
      InpEof = input->InpEof;
      real_input = input->save_real_input;
   }
}

void TEnvInput::CloseSubInput ()
{
   assert (input != NULL);

   #if 0
   if (opt.ShowExpansionDetails)
   {
      info ("closing sub input: " + QuoteStr (input->GetInfo ()));
      info ("next character: " + QuoteStr (CharToStr (input->extra_character)));
   }
   #endif

   if (input->is_include_file)
   {
      ClosingInclude ();
      if (opt.ShowIncludes)
         Info ("closing " + input->GetFileName ());
   }

   CleanSubInput ();
}

void TEnvInput::OpenInclude (string file_name, string orig_name)
{
   TSubInput * item = new TSubInput ();
   item->Open (file_name);

   item->is_include_file = true;
   item->orig_include_name = orig_name;
   item->add_new_line = true; // finish last line

   OpenSubInput (item);

   if (env != NULL)
      env->AddUsedFile (file_name);

   NextCh (); // read first character

   // !? !! SetStartToken (); // formal first token
}

void TEnvInput::OpenString (const string text, bool add_stop)
{
   TStrSubInput * item = new TStrSubInput (); // !? delete
   item->is_include_file = false;
   item->add_final_token = add_stop;
   item->value = text;

   if (input != NULL)
   {
      // copy location
      item->PositionInitialized = true;

      item->FileInx = input->FileInx;
      item->LinePos = input->LinePos;
      item->ColPos  = input->ColPos;
      item->CharPos = input->CharPos;

      item->TokenFileInx = input->FileInx;
      item->TokenLinePos = input->LinePos;
      item->TokenColPos  = input->ColPos;
      item->TokenCharPos = input->CharPos;
   }

   OpenSubInput (item);
   GetCh (); // skip current character (is stored as extra_character)
             // use GetCh, not NextCh ... character should not be stored into duplicated input

   // !? !! SetStartToken (); // formal first token
}

void TEnvInput::OpenMacro (TSubInput * item, bool add_stop)
{
   assert (input != NULL);
   assert (real_input != NULL);

   item->is_include_file = false;
   item->add_final_token = add_stop;

   OpenSubInput (item);

   GetCh (); // skip current character (is stored as extra_character)
             // use GetCh, not NextCh ... character should not be stored into duplicated input

   // !? !! SetStartToken (); // formal first token
}

/* ---------------------------------------------------------------------- */

TCondSymbol * TEnvInput::LocalSearchSymbol (const string name)
{
   TCondSymbol * symbol = NULL;

   if (input != NULL)
      symbol = input->local_symbols.SearchSymbol (name);

   if (symbol == NULL)
   {
      assert (env != NULL);
      symbol = env->SearchSymbol (name);
   }

   return symbol;
}

TCondSymbol * TEnvInput::MacroSearchSymbol (TStrSubInput * macro_input, const string name)
{
   TCondSymbol * symbol =
      macro_input->local_symbols.SearchSymbol (name);

   if (symbol == NULL)
      symbol = LocalSearchSymbol (name);

   return symbol;
}

string TEnvInput::ReadParameter (bool multi) // %%
{
   TTextStorage orig_stream; // original value for current parametr
   int level = 0; // level of nested parenthesis
   bool inside = false; // true => after some non-empty token
   bool tail = false; // true => add one space before next tokem

   // InpCh is first character of current parameter
   // TokenKind, TokenTxt is old token

   #if 0
   if (opt.ShowExpansionDetails)
      Info ("... begin parameter " + param->name + " = " +
            TokenKindToStr (TokenKind) + " " + QuoteStr (TokenTxt) +
            ", next character is " + QuoteStr (CharToStr (InpCh)) );
   #endif

   while (level > 0 || ! ( ( ! multi && InpCh == ',' ) || InpCh == ')') )
   {
      // now read first parameter token
      GetSimpleToken ();

      if (IsSeparator ('(')) level ++;
      if (IsSeparator (')')) level --;

      if (TokenKind == SpaceToken ||
          TokenKind == CommentToken ||
          TokenKind == LineContToken ||
          TokenKind == StartToken ||
          TokenKind == StopToken ||
          TokenKind == FinalToken ||
          TokenKind == EofToken)
      {
         if (inside) // skip leading spaces
            tail = true; // only one space
      }
      else
      {
         if (tail)
            orig_stream.put (' '); // previous space

         #if 0
         if (opt.ShowExpansionDetails)
            Info ("... raw parameter token " + TokenKindToStr (TokenKind) + " " + QuoteStr (TokenTxt) + ", level " + IntToStr (level));
         #endif

         orig_stream.put_str (TokenTxt); // collect parameter value

         tail = false;
         inside = true;
      }
   }

   #if 0
   if (opt.ShowExpansionDetails)
      Info ("... raw parameter value " + param->name + " = " + orig_stream.to_string ());
   #endif

   return orig_stream.to_string ();
}

string TEnvInput::ExpandParameter (const string orig_value) // %%
{
   if (dupl_input)
      AddRemark ("->");

   OpenString (orig_value, true);

   ResolveSymbol (); // skip macro name or right parenthesis (after opening sub-input)

   TTextStorage expanded_stream;

   while (TokenKind != StartToken &&
          TokenKind != StopToken &&
          TokenKind != FinalToken &&
          TokenKind != EofToken)
   {
      #if 0
      if (opt.ShowExpansionDetails)
         Info ("... collect parameter token " +
               TokenKindToStr (TokenKind) + " " + QuoteStr (TokenTxt) +
               ", input = " + input->GetInfo());
      #endif
      expanded_stream.put_str (TokenTxt);

      ResolveSymbol ();
   }

   CloseSubInput ();

   if (TokenKind == FinalToken)
   {
      GetSimpleToken (); // skip FinalToken
   }
   else
   {
      #if 0
      if (opt.ShowExpansionDetails)
         info ("on " + TokenKindToStr (TokenKind) + " " + QuoteStr (TokenTxt) + ", input = " + input->GetInfo());
      #endif
      ShowPositions ();
      Error ("FinalToken expected - internal bug");
   }

   return expanded_stream.to_string ();
}

void TEnvInput::ExpandParameters (TStrSubInput * macro_input, TCondSymbol * symbol) // %%
{
   GetSimpleToken (); // skip macro name, move to left parenthesis
   assert (IsSeparator ('('));
   // NO GetSimpleToken (); // keep left parenthesis

   TCondParam * param = symbol->first_param;
   bool finish = false;

   if (param == NULL)
      GetSimpleToken (); // skip left parenthesis, important

   #if 0
   if (opt.ShowExpansionDetails)
      if (param == NULL)
         Info ("macro with zero parameters " +  symbol->name);
      else
         Info ("macro with some parameters " +  symbol->name);
   #endif

   while (param != NULL && ! finish)
   {
      // InpCh is first character of current parameter
      // TokenKind, TokenTxt is old token

      // expand one parameter
      string orig_value = ReadParameter (param->multi);
      string expanded_value = ExpandParameter (orig_value);

      // store parameter value
      TCondSymbol * sym = macro_input->local_symbols.DefineNewSymbol (param->name, expanded_value);
      sym->local_param = true;
      sym->orig_value = orig_value;

      if (opt.ShowExpansionDetails)
         Info ("parameter " + symbol->name + "." + param->name + " = " + QuoteStr (expanded_value));

      param = param->next;

      // check comma (parameter separator)
      if (param != NULL)
      {
         if (param->multi && IsSeparator (')'))
         {
            // special case multi parameter, but without value
            TCondSymbol * sym = macro_input->local_symbols.DefineNewSymbol (param->name, "");
            sym->local_param = true;
            finish = true;  // end of loop
         }

         if (! finish && ! IsSeparator (','))
         {
            #if 0
            if (opt.ShowExpansionDetails)
               info ("on " + TokenKindToStr (TokenKind) + " " + QuoteStr (TokenTxt));
            #endif
            Error ("',' expected (in macro expansion of " + symbol->name + "." + param->name + ")");
         }
         // NO GetSimpleToken (); // keep comma as current token
      }
   }

   if (! IsSeparator (')'))
   {
      #if 0
      if (opt.ShowExpansionDetails)
         info ("on " + TokenKindToStr (TokenKind) + " " + QuoteStr (TokenTxt));
      #endif
      Error ("')' expected (in macro expansion of " + symbol->name + ")");
   }
   // NO GetSimpleToken (); // keep right parenthesis token on input
}

string TEnvInput::CompoundItem (TStrSubInput * macro_input, const string name)
{
   string result;

   TCondSymbol * symbol = macro_input->local_symbols.SearchSymbol (name);

   if (symbol == NULL)
      result = name;
   else if (symbol->local_param)
      result = symbol->orig_value;
   else
      result = symbol->value;

   if (opt.ShowExpansionDetails)
      Info ("compound item " + name + " --> " + result);

   return result;
}

void TEnvInput::StringSymbol (TStrSubInput * macro_input) // %%
{
   GetSimpleToken (); // skip hash mark

   if (TokenKind != IdentToken)
      Error ("Identifier expected after hash mark");

   string name = TokenTxt;
   TCondSymbol * symbol = MacroSearchSymbol (macro_input, name);

   if (symbol == NULL)
   {
      Error ("Undefined symbol after hash mark: " + name);
   }
   else
   {
      if (dupl_input)
         AddRemark ("~~>");

      string t = "";

      if (symbol->local_param)
         t = QuoteStr (symbol->orig_value); // use non-expanded parameter value
      else
         t = QuoteStr (symbol->value);

      OpenString (t, false);
      GetSimpleToken (); // skip original token

      if (opt.ShowExpansionDetails)
         Info ("string " + t);
   }
}

void TEnvInput::CompoundSymbol (TStrSubInput * macro_input) // %%
{
   GetSimpleToken ();

   // spaces around hash mark(s) are removed during macro definitions

   if (opt.HashDirectives && IsSeparator ('#'))
   {
      StringSymbol (macro_input);
   }
   else
   {
      // InpCh ... next character after current token

      if (CondValue &&
          TokenKind == IdentToken &&
          opt.HashDirectives &&
          InpCh == '#')
      {
         /* collect values from concatenated symbols */
         string compound_value = CompoundItem (macro_input, TokenTxt);

         while (InpCh == '#')
         {
            GetSimpleToken (); // read first hash mark

            if (InpCh != '#')
            {
               // Error ("Two hash marks expected: " + QuoteStr (CharToStr (InpCh)));
               StringSymbol (macro_input);
            }
            else
            {
                GetSimpleToken ();  // read second hash mark

                GetSimpleToken (); // read next token (identifier)

                // NO if (TokenKind != IdentToken)
                //    Error ("Identifier expected after hash marks: "+ QuoteStr (TokenTxt));
            }

            compound_value += CompoundItem (macro_input, TokenTxt);
         }

         if (dupl_input)
            AddRemark ("~~>");

         /* lookup compound symbol */
         TCondSymbol * symbol = LocalSearchSymbol (compound_value);

         if (symbol != NULL)
         {
            compound_value = symbol->value;
            if (opt.ShowExpansionDetails)
               Info ("compound symbol " + symbol->name + " --> " + symbol->value);
         }

         /* use collected value as new input */
         OpenString (compound_value, false);

         GetSimpleToken (); // skip previous identifier
                            // do it after OpenString, otherwise one token after compound name is lost
      }
   }
}

void TEnvInput::ExpandAgain (TCondSymbol * symbol, TStrSubInput * macro_input) // %%
{
   TTextStorage expanded_stream;

   OpenString (symbol->value, true);

   CompoundSymbol (macro_input); // skip right parenthesis (after OpenMacro)

   while (TokenKind != StartToken &&
          TokenKind != StopToken &&
          TokenKind != FinalToken &&
          TokenKind != EofToken)
   {
      TCondSymbol * sym = NULL;

      // search in local symbols
      if (TokenKind == IdentToken)
         sym = macro_input->local_symbols.SearchSymbol (TokenTxt);

      if (sym != NULL)
         expanded_stream.put_str (sym->value);
      else
         expanded_stream.put_str (TokenTxt);

      CompoundSymbol (macro_input);
   }

   CloseSubInput ();

   assert (TokenKind == FinalToken);
   // NO GetSimpleToken (); // keep FinalToken

   macro_input->value = expanded_stream.to_string (); // store pre-expanded text
}

void TEnvInput::ExpandSymbol (TCondSymbol * symbol) // %%
{
   if (opt.ShowMacroExpansion || opt.ShowExpansionDetails)
      Info ("use " + DefineToText (symbol));

   TStrSubInput * macro_input = new TStrSubInput (); // !? delete
   macro_input->title = symbol->name;
   macro_input->value = symbol->value; // value for "not function like" macros

   // copy location
   macro_input->PositionInitialized = true;

   macro_input->FileInx = symbol->FileInx;
   macro_input->LinePos = symbol->LinePos;
   macro_input->ColPos  = symbol->ColPos;
   macro_input->CharPos = symbol->CharPos;

   macro_input->TokenFileInx = symbol->FileInx;
   macro_input->TokenLinePos = symbol->LinePos;
   macro_input->TokenColPos  = symbol->ColPos;
   macro_input->TokenCharPos = symbol->CharPos;

   // without parameters, leave macro name as current token
   // with parameters, leave right parenthesis as current token
   // skip this token after sub-input is opened

   bool only_macro_name = false;

   // parameters
   if (symbol->with_params)
   {
      // skip spaces
      while (! InpEof && InpCh <= ' ' && InpCh != cr && InpCh != lf)
         NextCh ();

      if (InpCh != '(')
         // spaces after symbol name are ignored
         only_macro_name = true;
      else
         ExpandParameters (macro_input, symbol);
   }

   if (only_macro_name)
   {
      delete macro_input;
      macro_input = NULL;
   }
   else
   {
      if (dupl_input)
         AddRemark ("-->");

      // expand parameters and compund names in macro
      ExpandAgain (symbol, macro_input);

      if (dupl_input)
         AddRemark ("--->");

      if (opt.ShowMacroExpansion || opt.ShowExpansionDetails)
         Info ("expand " + macro_input->value);

      // expand macro
      OpenMacro (macro_input, false);

      ResolveSymbol (); // skip right parenthesis (after OpenMacro)
                        // use ResolveSymbol to expand following token
   }
}

void TEnvInput::ResolveSymbol () // %%
{
   // read macro identifier with parameters ( or read one another token )

   GetSimpleToken ();

   #if 0
   if (opt.ShowExpansionDetails)
      info ("ResolveSymbol: " + QuoteStr (TokenTxt));
   #endif

   if (TokenKind == IdentToken && CondValue)
   {
      // search symbol definition
      string symbol_name = TokenTxt;
      TCondSymbol * symbol = NULL;
      if (! opt.IgnoreMacroExpansion)
      {
         if (symbol_name == "__FILE__")
            SetStringToken (QuoteStr (RecallFileName (real_input->TokenFileInx)));
            // first_input - original source file_name
            // input - possible macro in another file
         else if (symbol_name == "__LINE__")
            SetNumberToken (IntToStr (real_input->TokenLinePos));
         else
            symbol = LocalSearchSymbol (symbol_name);
      }

      // check recursion
      if (symbol != NULL)
      {
         TSubInput * temp = input;
         while (temp != NULL)
         {
            if (symbol_name == temp->title)
               symbol = NULL; // no recursion allowed
            temp = temp->prev;
         }
      }

      // read parameters and expand macro
      if (symbol != NULL)
         ExpandSymbol (symbol);
   }

   #if 0
      if (opt.ShowExpansionDetails)
         info ("ResolveSymbol next character " + QuoteStr (CharToStr (InpCh)));
   #endif
}

/* ---------------------------------------------------------------------- */

void TEnvInput::SkipSpaces ()
{
   while (TokenKind == SpaceToken ||
          TokenKind == CommentToken ||
          TokenKind == LineContToken)
   {
      GetSimpleToken ();
   }
}

void TEnvInput::CheckEndOfLine ()
{
   SkipSpaces ();

   if (! opt.TolerateMissingEndOfLine)
      if (TokenKind != EolToken &&
          TokenKind != StopToken &&
          TokenKind != EofToken)
      {
         Error ("End of line (after directive) expected"
                ", TokenTxt = " + TokenTxt +
                ", TokenKind = " + TokenKindToStr (TokenKind));
      }
}

/********************************** DEFINE ********************************/

bool TEnvInput::ThreeDots ()
{
   bool result;
   if (compound_separators)
   {
       result = IsSeparator() && TokenTxt == "...";
       if (result)
          GetSimpleToken ();
   }
   else
   {
      result = IsSeparator ('.');

      if (result)
      {
         GetSimpleToken ();

         if (! IsSeparator ('.'))
            Error ("'...' expected");

         GetSimpleToken ();

         if (! IsSeparator ('.'))
            Error ("'...' expected");

         GetSimpleToken ();
      }
   }
   return result;
}

void TEnvInput::DefineParameter (TCondSymbol * symbol)
{
   string name = "";
   bool multi = false;

   if (ThreeDots ())
   {
       name = "__VA_ARGS__";
       multi = true;
   }
   else
   {
      if (TokenKind != IdentToken)
         Error ("Parameter identifier expected");

      name = TokenTxt;
      GetSimpleToken (); // skip parameter name
      SkipSpaces ();

      if (ThreeDots ())
      {
         multi = true;
         SkipSpaces ();
      }
   }

   // debug ("DefineParameter " + name);
   symbol->InsParam (name, multi);
}

string TEnvInput::DefineToText (TCondSymbol * symbol)
{
   string txt = symbol->name;

   if (symbol->with_params)
   {
      txt += "(";
      TCondParam * param = symbol->first_param;
      while (param != NULL)
      {
         if (param->multi)
            txt += "...";
         else
            txt += param->name;

         param = param->next;

         if (param != NULL)
            txt += ",";
      }
      txt += ") ";
   }

   txt += + " --> " + symbol->value;

   return txt;
}

void TEnvInput::DefineDirective ()
{
   int cont_count = 0;
   GetSimpleToken (); // skip define
   SkipSpaces ();

   if (TokenKind != IdentToken)
      Error ("Identifier expected");

   TCondSymbol * symbol = new TCondSymbol ();
   symbol->name = TokenTxt;

   // debug ("DefineDirective " + TokenTxt);

   symbol->with_params = (InpCh == '('); // parenthesis without space
   GetSimpleToken (); // skip symbol name
   SkipSpaces ();

   if (symbol->with_params)
   {
      GetSimpleToken (); // skip '('
      SkipSpaces ();

      if (! IsSeparator (')'))
      {
         DefineParameter (symbol);
         while (IsSeparator (','))
         {
            if (symbol->last_param->multi)
               Error ("'...' must be last parameter");
            GetSimpleToken ();
            SkipSpaces ();
            DefineParameter (symbol);
         }
      }

      if (! IsSeparator (')'))
         Error ("')' expected (in macro definition))");

      GetSimpleToken ();
      SkipSpaces ();
   }

   assert (input != NULL);
   symbol->FileName = input->GetFileName ();
   symbol->FileInx = input->FileInx;
   symbol->LinePos = input->LinePos;
   symbol->ColPos  = input->ColPos;
   symbol->CharPos = input->CharPos;

   // collect value, trim white spaces
   bool inside = false;
   bool space = false;
   bool ignore_space = false;
   string value = "";

   while (TokenKind != EolToken &&
          TokenKind != StopToken &&
          TokenKind != FinalToken &&
          TokenKind != EofToken)
   {
      if (TokenKind == LineContToken)
         cont_count ++;

      if (TokenKind == SpaceToken || TokenKind == CommentToken || TokenKind == LineContToken)
      {
         if (inside) // skip leading spaces
            space = true;
      }
      else
      {
         if (TokenTxt == "#") // !?
            ignore_space = true; // ignore space before ##

         if (space && ! ignore_space)
            value += " ";

         value += TokenTxt;

         inside = true;
         space = false;
         ignore_space = (TokenTxt == "#" || TokenTxt == "##"); // ignore space after # or ##
      }

      // skip trailing space
      GetSimpleToken ();
   }

   symbol->value = value;

   if (opt.ShowDefinitions)
      Info ("define " + DefineToText (symbol));

   if (CondValue) // is it valid source section
   {
      assert (input != NULL);
      env->UndefineSymbol (symbol->name);
      env->InsSymbol (symbol);
   }
   else
   {
      delete symbol; // !?
   }
}

void TEnvInput::UndefDirective ()
{
   GetSimpleToken (); // skip undef
   SkipSpaces ();

   if (TokenKind != IdentToken)
      Error ("Identifier expected");

   if (CondValue) // is it valid source section
      env->UndefineSymbol (TokenTxt);

   GetSimpleToken (); // skip symbol name
   CheckEndOfLine ();
}

/****************************** EXPRESSIONS *******************************/

void TEnvInput::GetExprToken () // %%
{
   again:

   ResolveSymbol ();

   if (TokenKind == LineContToken) line_cont ++;

   if (TokenKind == StartToken) goto again;
   if (TokenKind == StopToken) goto again;
   if (TokenKind == CommentToken) goto again;
   if (TokenKind == LineContToken) goto again;
   if (TokenKind == SpaceToken) goto again;
}

/* ---------------------------------------------------------------------- */

bool TEnvInput::TwoSeparators (char c1, char c2)
{
    bool result;
    if (compound_separators)
    {
       result = IsSeparator () &&
                TokenTxt.length () == 2 &&
                TokenTxt [0] == c1 &&
                TokenTxt [1] == c2;
    }
    else
    {
       result = IsSeparator (c1) && (InpCh == c2);
    }
    return result;
}

void TEnvInput::SkipTwoSeparators ()
{
   GetExprToken ();
   if (! compound_separators)
      GetExprToken ();
}

bool TEnvInput::CondSymbolDefined ()
{
   if (TokenKind != IdentToken || TokenTxt == "defined")
      Error ("Identifier of conditional symbol expected");

   // search identifier
   TCondSymbol * symbol = LocalSearchSymbol (TokenTxt);
   bool result = (symbol != NULL);

   GetExprToken (); // skip identifier

   return result;
}

int TEnvInput::CondIdent (bool cont)
{
   if (TokenKind != IdentToken || TokenTxt == "defined")
      Error ("Identifier of conditional symbol expected");

   // search identifier
   TCondSymbol * symbol = LocalSearchSymbol (TokenTxt);

   // !? !! BUG
   if (symbol != NULL && IsValidIdent (symbol->value))
      symbol = LocalSearchSymbol (symbol->value);

   int result = 0;

   if (symbol == NULL)
   {
      if (opt.RequireSymbolDefinition && cont)
         Error ("Unknown conditional symbol: " + TokenTxt);
         // cont == false => symbol need not exist
   }
   else
   {
      bool ok;
      StrToInt (symbol->value, result, ok);
      if (! ok && opt.RequireSymbolDefinition && cont)
         Error ("Numeric conditional symbol expected: " + TokenTxt + " (" + symbol->value + ")");
   }

   GetExprToken (); // skip identifier

   if (! cont)
   {
      // macro parameters of possible undefined macro identifier
      if (IsSeparator ('('))
      {
         GetExprToken (); // skip left parenthesis

         int level = 0;
         while ( level > 0 || ! IsSeparator (')') )
         {
            if (IsSeparator ('(')) level ++;
            if (IsSeparator (')')) level --;

            GetExprToken (); // next token
         }

         if (! IsSeparator (')'))
            Error ("')' expected");

         GetExprToken (); // skip right parenthesis
      }
   }

   return result;
}

/* ---------------------------------------------------------------------- */

int TEnvInput::CondSimpleExpr (bool cont)
{
   int result = 0;

   if (TokenKind == NumberToken)
   {
      string txt = TokenTxt;
      int len = txt.length ();
      int inx = len;

      while (inx > 1 &&
             (txt [inx-1] == 'l' ||
              txt [inx-1] == 'L' ||
              txt [inx-1] == 'u' ||
              txt [inx-1] == 'U' ))
      {
         inx--;
      }

      if (inx < len)
         txt = Head (txt, inx);

      bool ok = StrToInt (txt, result);
      if (! ok)
      {
         unsigned long temp;
         ok = StrToNum (txt, temp);
         if (ok)
            result = temp;
         else
            Error ("Invalid number: " + txt);
      }
      GetExprToken (); // skip number
   }
   else if (TokenKind == IdentToken)
   {
      if (TokenTxt == "defined")
      {
         GetSimpleToken (); // skip defined, no macro expansion
         SkipSpaces ();

         if (IsSeparator ('('))
         {
            GetSimpleToken (); // skip parenthesis, no macro expansion
            SkipSpaces ();

            result = CondSymbolDefined ();
            if (! IsSeparator (')'))
               Error ("')' expected (in defined)");

            GetExprToken (); // skip parenthesis, expand symbol
         }
         else
         {
            result = CondSymbolDefined ();
         }
      }
      else
      {
          result = CondIdent (cont);

          #if 0
          if (! cont) // read possible macro arguments
          {
             if (IsSeparator ('('))
             {
                int level = 1;

                while (level > 0)
                {
                   GetSimpleToken (); // skip parenthesis, no macro expansion

                   if (IsSeparator ('('))
                      level ++;

                   if (IsSeparator (')'))
                      level --;
                }

                GetExprToken (); // skip parenthesis, expand symbol
             }
          }
          #endif
      }
   }
   else if (IsSeparator ('!'))
   {
      GetExprToken (); // skip exclamation mark
      result = ! CondSimpleExpr (cont);
   }
   else if (IsSeparator ('~'))
   {
      GetExprToken (); // skip exclamation mark
      result = ~ CondSimpleExpr (cont);
   }
   else if (IsSeparator ('('))
   {
      GetExprToken (); // skip left parenthesis
      result = CondExpr (cont);
      if (! IsSeparator (')'))
         Error ("')' expected (in subexpression): " + TokenTxt);
      GetExprToken (); // skip right parenthesis
   }
   else Error ("Simple conditional expression expected: " + TokenTxt);
                // ", TokenTxt = " + TokenTxt +
                // ", TokenKind = " + TokenKindToStr (TokenKind));

   return result;
}

int TEnvInput::CondFactor (bool cont)
{
   int result = CondSimpleExpr (cont);

   while (IsSeparator ('*') || IsSeparator ('/') || IsSeparator ('%'))
   {
      char mark = TokenSym;
      GetExprToken ();

      int temp = CondSimpleExpr (cont);

      // !?
      if (mark == '*')
      {
         result = result * temp;
      }
      else
      {
         if (temp == 0)
            Error ("Zero divide");

         if (mark == '/')
            result = result / temp;
         else
            result = result % temp;
      }
   }

   return result;
}

int TEnvInput::CondTerm (bool cont)
{
   int result = CondFactor (cont);

   while (IsSeparator ('+') || IsSeparator ('-'))
   {
      char mark = TokenSym;
      GetExprToken ();

      int temp = CondFactor (cont);

      if (mark == '+')
         result = result + temp;
      else
         result = result - temp;
   }

   return result;
}

int TEnvInput::CondShift (bool cont)
{
   int result = CondTerm (cont);

   while (TwoSeparators ('<', '<') || TwoSeparators ('>', '>'))
   {
      char left = TwoSeparators ('<', '<');
      SkipTwoSeparators ();

      int temp = CondTerm (cont);

      if (left)
         result = result << temp;
      else
         result = result >> temp;
   }

   return result;
}

int TEnvInput::CondRel (bool cont)
{
   int result = CondShift (cont);

   while (IsSeparator ('<') || IsSeparator ('>'))
   {
      char mark = TokenSym;
      GetExprToken (); // skip first

      char mark2 = ' ';
      if (IsSeparator ('='))
      {
         mark2 = TokenSym;
         GetExprToken (); // skip second
      }

      int temp = CondShift (cont);

      if (mark == '<')
      {
         if (mark2 == '=')
            result = (result <= temp);
         else
            result = (result < temp);
      }
      else
      {
         if (mark2 == '=')
            result = (result >= temp);
         else
            result = (result > temp);
      }
   }

   return result;
}

int TEnvInput::CondEq (bool cont)
{
   int result = CondRel (cont);

   while (TwoSeparators ('=', '=') || TwoSeparators ('!', '=') )
   {
      bool eq = TwoSeparators ('=', '=');
      SkipTwoSeparators ();

      int temp = CondRel (cont);

      if (eq)
         result = (result == temp);
      else
         result = (result != temp);
   }

   return result;
}

int TEnvInput::CondAnd (bool cont)
{
   int result = CondEq (cont);

   while (IsSeparator ('&') && ! TwoSeparators ('&', '&'))
   {
      GetExprToken (); // skip &

      int temp = CondEq (cont);
      result = result & temp;
   }

   return result;
}

int TEnvInput::CondXor (bool cont)
{
   int result = CondAnd (cont);

   while (IsSeparator ('^'))
   {
      GetExprToken (); // skip ^

      int temp = CondAnd (cont);
      result = result ^ temp;
   }

   return result;
}

int TEnvInput::CondOr (bool cont)
{
   int result = CondXor (cont);

   while (IsSeparator ('|') && ! TwoSeparators ('|', '|'))
   {
      GetExprToken (); // skip |

      int temp = CondXor (cont);
      result = result | temp;
   }

   return result;
}

int TEnvInput::CondLogAnd (bool cont)
{
   int result = CondOr (cont);

   while (TwoSeparators ('&', '&'))
   {
      SkipTwoSeparators ();

      int temp = CondOr (cont && result);
      result = result && temp;
   }

   return result;
}

int TEnvInput::CondLogOr (bool cont)
{
   int result = CondLogAnd (cont);

   while (TwoSeparators ('|', '|'))
   {
      SkipTwoSeparators ();

      int temp = CondLogAnd (cont && ! result);
      result = result || temp;
   }

   return result;
}

int TEnvInput::CondExpr (bool cont)
{
   int result = CondLogOr (cont);

   if (IsSeparator ('?'))
   {
      // Debug ("QUESTION mark");
      GetExprToken (); // skip

      // GetSimpleToken();
      int branch1 = CondLogOr (cont && result);

      // DO NOT USE CheckSeparator (which uses GetToken)
      if (! IsSeparator (':'))
         Error ("Colon expected");

      // Debug ("COLON");
      GetExprToken (); // skip

      int branch2 = CondLogOr (cont && ! result);

      result = result ? branch1 : branch2;
   }

   return result;
}

/************************* CONDITIONAL DIRECTIVES *************************/

void TEnvInput::CondDirective (bool inv, bool only_symbol)
{
   bool value;

   if (only_symbol)
   {
      GetSimpleToken (); // skip if, no macro expansion
      SkipSpaces ();

      if (real_input->guard_state == open_guard)
      {
         // second part of first ifndef
         real_input->guard_name = TokenTxt; // remember variable name
         real_input->guard_state = inside_guard;
      }

      value = CondSymbolDefined ();
   }
   else
   {
      GetExprToken (); // skip if and expand next symbol
      value = CondExpr (CondValue); // be tolerant to unknown symbols and macros, when CondValue == false
   }

   if (inv)
      value = ! value;

   if (CondTable.level+1 > cond_level_max)
      Error ("Too many nested conditional directives");

   CondTable.AddCondLevel (value);

   CheckEndOfLine ();
}

/* ---------------------------------------------------------------------- */

void TEnvInput::IfDirective ()
{
   CondDirective (false, false); // CondDirective (inversion, only_symbol)
}

void TEnvInput::IfdefDirective ()
{
   CondDirective (false, true);
}

void TEnvInput::IfndefDirective ()
{
   if (real_input->guard_state == before_guard)
   {
      // first ifndef encountered
      real_input->guard_state = open_guard;
      real_input->guard_level = CondTable.level + 1;
   }

   CondDirective (true, true);
}

void TEnvInput::ElifDirective ()
{
   if (real_input->guard_state == inside_guard &&
       CondTable.level == real_input->guard_level)
   {
      real_input->guard_state = wrong_guard;
      // first ifndef cannot have elif
   }

   if (CondTable.current_level == NULL)
      Error ("elif directive without if directive");

   if (CondTable.current_level->value)
      CondTable.current_level->done = true;

   if (CondTable.current_level->done)
   {
      // only read condition and skip following conditional section
      // Debug ("Before reading expression");
      GetExprToken (); // skip elif and expand next symbol
      CondExpr (false);
      CondTable.current_level->value = false;
      // Debug ("After reading expression");
   }
   else
   {
      // Debug ("Before expression");
      CondValue = true; // importat, turn on conditional compilation (before reading first macro symbol)
      GetExprToken (); // skip elif and expand next symbol
      CondTable.current_level->value = CondExpr ();
      // Debug ("After expression");
   }

   CheckEndOfLine ();
}

void TEnvInput::ElseDirective ()
{
   if (real_input->guard_state == inside_guard &&
      CondTable.level == real_input->guard_level)
   {
      real_input->guard_state = wrong_guard;
      // first ifndef cannot have else
   }

   GetSimpleToken (); // skip else

   if (CondTable.current_level == NULL)
      Error ("else directive without if directive");

   if (CondTable.current_level->value)
      CondTable.current_level->done = true;

   CondTable.current_level->value = ! CondTable.current_level->done;

   CheckEndOfLine ();
}

void TEnvInput::EndifDirective ()
{
   if (real_input->guard_state == inside_guard &&
       CondTable.level == real_input->guard_level)
   {
      real_input->guard_state = close_guard;
      // last endif encountered
   }

   GetSimpleToken (); // skip endif

   if (CondTable.current_level == NULL)
      Error ("endif directive without if directive");
   else
      CondTable.RemoveCondLevel ();

   CheckEndOfLine ();
}

/******************************** INCLUDE *********************************/

void TEnvInput::IncludeDirective (bool include_next) // %%
{
   if (! CondValue)
   {
      GetSimpleToken (); // skip include
      SkipSpaces ();
   }

   if (CondValue) // valid source section
   {
      // skip spaces
      while (! InpEof && InpCh <= ' ' && InpCh != cr && InpCh != lf)
         NextCh ();

      bool simple = (InpCh == quote2 || InpCh == '<');

      // simple == false => expand macro names
      // #define _MULTIARCH_HEADER(arch,name) <arch-name>
      // #include _MULTIARCH_HEADER(i386,name)

      // simple == true => do not expand macro names within < >
      // #define linux 1
      // #include <linux/limits.h>

      if (simple)
         GetSimpleToken ();
      else
         GetExprToken (); // skip include and expand next symbol

      // check only if enabled input, important for <file name>

      string symbolic_name = "";
      string name = "";
      bool   system_dir = false;

      if (TokenKind == StringToken)
      {
         symbolic_name = TokenTxt;
         name = TokenVal;
      }
      else if (IsSeparator ('<'))
      {
         if (simple)
            GetSimpleToken ();
         else
            GetExprToken (); // skip '<'

         while (TokenKind != EolToken &&
                TokenKind != StopToken &&
                TokenKind != FinalToken &&
                TokenKind != EofToken &&
                ! IsSeparator ('>'))
         {
            name = name + TokenTxt;
            if (simple)
               GetSimpleToken ();
            else
               GetExprToken (); // store character also into input_storage
         }

         if (! IsSeparator ('>'))
            Error ("'>' expected");

         symbolic_name = '<' + name + '>';
         system_dir = true;
      }
      else
      {
         Error ("File name expected: " + TokenTxt);
      }

      GetSimpleToken ();
      CheckEndOfLine (); // check end of line, but do not move to another line

      if (opt.ShowIncludes)
         Info ("opening " + symbolic_name);

      string orig_name = name;

      if (env != NULL && ! opt.IgnoreIncludes)
      {
         assert (input != NULL);
         assert (real_input != NULL);
         string dir = ExtractPath (real_input->GetFileName ());

         name = env->TranslatePath (name);
         if (! include_next)
         {
            // Debug ("Searching " + name);
            name = env->LookupInclFile (name, dir, system_dir);
         }
         else
         {
            string this_file_name = input->orig_include_name;
            string next_file_name = name;
            if (opt.ShowInclNext)
               Info ("include_next " + this_file_name + " and then " + next_file_name);
            name = env->LookupInclFile (this_file_name, dir, system_dir, next_file_name, opt.ShowInclNext);
            if (opt.ShowInclNext && name != "")
               info ("found " + name);
         }

         if (name == "" && ! opt.TolerateIncludes)
            Error ("Cannot open include file: " + symbolic_name);
      }
      else
      {
         name = "";
      }

      if (name == "")
      {
         UnknownInclude (symbolic_name);
      }
      else if (opt.ReportIncludes)
      {
         // report include directive (without processing)
         SetIncludeDirectiveToken (name); // (with translated file name)
      }
      else
      {
         name = QueryInclude (name); // optional file name transformation

         // empty name => skip include file
         // ( used for already processed files )

         if (name != "")
            OpenInclude (name, orig_name);
      }
   }
}

/**************************** OTHER DIRECTIVES ****************************/

void TEnvInput::PragmaDirective ()
{
   // string name1 = TokenTxt;
   GetSimpleToken (); // skip pragma
   SkipSpaces ();

   string name2 = TokenTxt;
   // no GetInnerToken (); // skip name

   if (! CustomPragma (name2))
   {
      // read remaining text
      string txt;
      while (TokenKind != EolToken &&
             TokenKind != StopToken &&
             TokenKind != FinalToken &&
             TokenKind != EofToken)
      {
         txt = txt + TokenTxt;
         GetSimpleToken ();
      }
   }
}

/* ---------------------------------------------------------------------- */

void TEnvInput::WarningDirective ()
{
   GetSimpleToken (); // skip name
   SkipSpaces ();

   string txt;

   while (TokenKind != EolToken &&
          TokenKind != StopToken &&
          TokenKind != FinalToken &&
          TokenKind != EofToken)
   {
      txt = txt + TokenTxt;
      GetSimpleToken ();
   }

   if (CondValue && ! opt.IgnoreWarningDirectives)
      Warning (txt);
}

void TEnvInput::ErrorDirective ()
{
   GetSimpleToken (); // skip name
   SkipSpaces ();

   string txt;

   while (TokenKind != EolToken && TokenKind != StopToken && TokenKind != EofToken)
   {
      txt = txt + TokenTxt;
      GetSimpleToken ();
   }

   if (CondValue && ! opt.IgnoreErrorDirectives)
      Error (txt);
}

/* ---------------------------------------------------------------------- */

void TEnvInput::ExtraCh ()
{
   TokenTxt += InpCh; // !?
   TokenSym = 0; // multi character

   #ifdef DUPL_OUTPUT
      if (dupl_output)
         output_storage.put (InpCh);
   #endif

   NextCh ();
}

string TEnvInput::GetGuardSymbol ()
{
   // !? !!
   if (real_input->guard_state == after_guard)
      return real_input->guard_name;
   else
      return "";
}

/******************************* GET TOKEN ********************************/

void TEnvInput::GetToken ()
{
   again:

   ResolveSymbol (); // read one token

   string name = "";

   if ((IsSeparator ('#') && opt.HashDirectives && ! TwoSeparators ('#', '#')) || // !?
       (IsSeparator ('@') && opt.AtDirectives) ||
       (IsSeparator ('$') && opt.DollarDirectives))
   {
      assert (input != NULL);
      int SaveFileInx = input->TokenFileInx;
      int SaveLinePos = input->TokenLinePos;
      int SaveColPos  = input->TokenColPos;
      int SaveCharPos = input->TokenCharPos;

      if (opt.KeepUnusedSections)
      {
         dupl_directives = true;
         directive_storage.put_str (TokenTxt);
      }

      line_cont = 0;

      GetSimpleToken (); // skip hash mark ( and comments )
      SkipSpaces ();

      if (TokenKind != EolToken)
      {
         if (TokenKind != IdentToken)
            Error ("Directive name expected");

         name = TokenTxt;

         if (name == "define")
            DefineDirective ();

         else if (name == "undef")
            UndefDirective ();

         else if (name == "if")
            IfDirective ();

         else if (name == "ifdef")
            IfdefDirective ();

         else if (name == "ifndef")
            IfndefDirective ();

         else if (name == "elif")
            ElifDirective ();

         else if (name == "else")
            ElseDirective ();

         else if (name == "endif")
            EndifDirective ();

         else if (name == "include")
            IncludeDirective ();

         else if (name == "include_next")
            IncludeDirective (true);

         else if (name == "pragma")
            PragmaDirective ();

         else if (name == "warning")
            WarningDirective ();

         else if (name == "error")
            ErrorDirective ();

         else if (CustomDirective (name))
            /* nothing */ ;

         else if (! opt.IgnoreUnknownDirectives)
               Error ("Unknown directive: " + name);
      }

      string directive_text = "";
      if (opt.KeepUnusedSections)
      {
         directive_text = directive_storage.to_string ();
         directive_storage.reset ();
         dupl_directives = false;
      }

      assert (input != NULL);
      input->TokenFileInx = SaveFileInx;
      input->TokenLinePos = SaveLinePos;
      input->TokenColPos  = SaveColPos;
      input->TokenCharPos = SaveCharPos;

      // calculate value
      CondValue = CondTable.Calculate ();

      // set TokenKind, ... (if not already assigned)
      if (TokenKind != IncludeDirectiveToken)
      {
          if (opt.KeepUnusedSections)
          {
             SetUnusedDirectiveToken (directive_text);
          }
          else if (opt.KeepEmptySections)
          {
             string txt = "";
             for (int i = 1; i <= line_cont+1; i++)
                txt += lf;
             SetUnusedDirectiveToken (txt);
          }
          else
          {
             SetProcessedDirectiveToken ();
          }
      }
   }

   // ifdef ... endif guard
   if (real_input->guard_state == before_guard ||
       real_input->guard_state == after_guard)
   {
      if (TokenKind != StartToken &&
          TokenKind != CommentToken &&
          TokenKind != LineContToken &&
          TokenKind != SpaceToken &&
          TokenKind != UnusedDirectiveToken &&
          TokenKind != EolToken &&
          TokenKind != StopToken &&
          TokenKind != FinalToken &&
          TokenKind != EofToken)
      {
         real_input->guard_state = wrong_guard;
         // only "empty" tokens alowed before first ifdef or after last endif
      }
   }

   if (real_input->guard_state == close_guard)
   {
      real_input->guard_state = after_guard; // after last endif
   }

   // evaluate condition
   if (! CondValue)
   {
      if (TokenKind == EofToken)
         Error ("Unterminated conditional directive");

      if (! opt.KeepUnusedSections)
         if (! opt.KeepEmptySections || TokenKind != EolToken)
            goto again;

      TokenKind = UnusedDirectiveToken; // change token type
   }

   if (TokenKind == StartToken)     goto again;
   if (TokenKind == StopToken)      goto again;
   if (TokenKind == ProcessedDirectiveToken) goto again;

   #ifdef DUPL_OUTPUT
      if (dupl_output)
      {
         #if 0
         output_storage.put_str (QuoteStr(TokenTxt));
         if (TokenTxt == "\n")
            output_storage.put ('\n');
         else
            output_storage.put (' ');
         #endif

         output_storage.put_str (TokenTxt);
      }
   #endif

   if (TokenKind == SpaceToken    && ! opt.ReportSpaces)   goto again;
   if (TokenKind == CommentToken  && ! opt.ReportComments) goto again;
   if (TokenKind == LineContToken && ! opt.ReportLineCont) goto again;
   if (TokenKind == EolToken      && ! opt.ReportEol)      goto again;

   if (opt.ShowTokens)
      Info ("token: " + TokenKindToStr (TokenKind) + " " + QuoteStr (TokenTxt));
}

/* ---------------------------------------------------------------------- */

CLOSE_NAMESPACE
